import networkx as nx
import re

from enum import Enum

import requests
import sys

class TermState:
    ANY = 1
    VALID = 2
    OBSOLETED = 3
    MERGED = 4

def value(var):
    return var if var is not None else "N/A"
    
def relationValue(array):
    if array is None:
        return "N/A"
    
    
    
class NamedEntity:
    
    def __init__(self):
        self.id = None
        self.label = None
    
    
class Term:
    
    def __init__(self):
        self.id = None
        self.alt_ids = None
        self.is_obsolete = False
        self.is_a = None
        self.namespace = None
        self.name = None
        self.comment = None
        self.synonyms = None
        self.definition = None
        self.created_by = None
        self.creation_date = None
        self.subsets = None
        self.xrefs = None
        self.intersection_of = None
        self.relationship = None
        
        
    def count_metas(self, includeXRefs = True):
        count = 0
        if self.id:
            count += 1
        if self.alt_ids:
            count += len(self.alt_ids)
        # if self.is_obsolete:
        #     count += 1
        if self.namespace:
            count += 1
        if self.name:
            count += 1
        if self.comment:
            count += 1
        if self.synonyms:
            count += len(self.synonyms)
        if self.definition:
            count += 1
        if self.subsets:
            count += len(self.subsets)
        if self.xrefs and includeXRefs:
            count += len(self.xrefs)
        return count
        
    def count_structurals(self):
        count = 0
        if self.is_a:
            count += len(self.is_a)
        if self.relationship:
            count += len(self.relationship)
        if self.intersection_of:
            count == len(self.intersection_of)
        return count
        

    def is_merged(self):
        return self.alt_ids is not None
        
        
    def has_alt_id(self, query):
        if self.alt_ids is None:
            return False
        for id in self.alt_ids:
            if id == query:
                return True
        return False
        
        
    def add_intersection_of(self, relationship, target_id, target_label):
        if self.intersection_of is None:
            self.intersection_of = []

        relation = Relation()
        if relationship == "":
            relation.name = "is_a"
        else:
            relation.name = relationship

#        print("relation: " + relation.name + " id: " + target_id + " label: " + target_label)

        target = NamedEntity()
        target.id = target_id
        target.label = target_label
        
        relationTarget = RelationTarget()
        relationTarget.relation = relation
        relationTarget.target = target
        
#        self.intersection_of.append(relationTarget)
        self.intersection_of.append(relation.name + " " + target_id)
        
    def add_relationship(self, relationship, target_id, target_label):
        if self.relationship is None:
            self.relationship = []

        relation = Relation()
        if relationship == "":
            relation.name = "is_a"
        else:
            relation.name = relationship

#        print("relation: " + relation.name + " id: " + target_id + " label: " + target_label)

        target = NamedEntity()
        target.id = target_id
        target.label = target_label
        
        relationTarget = RelationTarget()
        relationTarget.relation = relation
        relationTarget.target = target
        
#        self.relationship.append(relationTarget)
        self.relationship.append(relation.name + " " + target_id)
        
        
    def add_is_a(self, is_a):
        if self.is_a is None:
            self.is_a = []
        self.is_a.append(is_a)
        
    def add_alternate_id(self, alt_id):
        if self.alt_ids is None:
            self.alt_ids = []
        self.alt_ids.append(alt_id)
        
    def add_synonym(self, synonym):
        if self.synonyms is None:
            self.synonyms = []
        self.synonyms.append(synonym)
        
    def add_subset(self, subset):
        if self.subsets is None:
            self.subsets = []
        self.subsets.append(subset)
        
    def add_xref(self, xref):
        if self.xrefs is None:
            self.xrefs = []
        self.xrefs.append(xref)

    def equals(self, other):
        return self.id == other.id and self.is_obsolete == other.is_obsolete and self.alt_ids == other.alt_ids and self.name == other.name and self.is_a == other.is_a and self.namespace == other.namespace and self.definition == other.definition and self.comment == other.comment and self.synonyms == other.synonyms and self.subsets == other.subsets and self.xrefs == other.xrefs and self.relationship == other.relationship and self.intersection_of == other.intersection_of

    def structural_equals(self, other):
        return self.is_a == other.is_a and self.relationship == other.relationship and self.intersection_of == other.intersection_of

    def meta_equals(self, other, includeXRefs = True):
        if includeXRefs:
            return self.id == other.id and self.is_obsolete == other.is_obsolete and self.alt_ids == other.alt_ids and self.name == other.name and self.namespace == other.namespace and self.definition == other.definition and self.comment == other.comment and self.synonyms == other.synonyms and self.subsets == other.subsets and self.xrefs == other.xrefs
        return self.id == other.id and self.is_obsolete == other.is_obsolete and self.alt_ids == other.alt_ids and self.name == other.name and self.namespace == other.namespace and self.definition == other.definition and self.comment == other.comment and self.synonyms == other.synonyms and self.subsets == other.subsets

    def explain_structural_differences(self, other):
        reasons = {}
        if self.is_a != other.is_a:
            reasons["is_a"] = {"current": value(self.is_a), "previous": value(other.is_a) }
        if self.relationship != other.relationship:
            reasons["relationship"] = {"current": value(self.relationship), "previous": value(other.relationship) }
        if self.intersection_of != other.intersection_of:
            reasons["intersection_of"] = {"current": value(self.intersection_of), "previous": value(other.intersection_of) }
        return reasons

    def explain_meta_differences(self, other, includeXRefs = True):
        reasons = {}
        if self.id != other.id:
            reasons["id"] = {"current": value(self.id), "previous": value(other.id) }
        if self.is_obsolete != other.is_obsolete:
            reasons["is_obsolete"] = {"current": value(self.is_obsolete), "previous": value(other.is_obsolete) }
        if self.alt_ids != other.alt_ids:
            reasons["alt_ids"] = {"current": value(self.alt_ids), "previous": value(other.alt_ids) }
        if self.name != other.name:
            reasons["name"] = {"current": value(self.name), "previous": value(other.name) }
        if self.namespace != other.namespace:
            reasons["namespace"] = {"current": value(self.namespace), "previous": value(other.namespace) }
        if self.definition != other.definition:
            reasons["definition"] = {"current": value(self.definition), "previous": value(other.definition) }
        if self.comment != other.comment:
            reasons["comment"] = {"current": value(self.comment), "previous": value(other.comment) }
        if self.synonyms != other.synonyms:
            reasons["synonyms"] = {"current": value(self.synonyms), "previous": value(other.synonyms) }
        if self.subsets != other.subsets:
            reasons["subsets"] = {"current": value(self.subsets), "previous": value(other.subsets) }
        if self.xrefs != other.xrefs and includeXRefs:
            reasons["xrefs"] = {"current": value(self.xrefs), "previous": value(other.xrefs) }
        return reasons
        
    def explain_differences(self, other):
        reasons = {}
        if self.id != other.id:
            reasons["id"] = {"current": value(self.id), "previous": value(other.id) }
        if self.is_obsolete != other.is_obsolete:
            reasons["is_obsolete"] = {"current": value(self.is_obsolete), "previous": value(other.is_obsolete) }
        if self.alt_ids != other.alt_ids:
            reasons["alt_ids"] = {"current": value(self.alt_ids), "previous": value(other.alt_ids) }
        if self.name != other.name:
            reasons["name"] = {"current": value(self.name), "previous": value(other.name) }
        if self.namespace != other.namespace:
            reasons["namespace"] = {"current": value(self.namespace), "previous": value(other.namespace) }
        if self.definition != other.definition:
            reasons["definition"] = {"current": value(self.definition), "previous": value(other.definition) }
        if self.comment != other.comment:
            reasons["comment"] = {"current": value(self.comment), "previous": value(other.comment) }
        if self.synonyms != other.synonyms:
            reasons["synonyms"] = {"current": value(self.synonyms), "previous": value(other.synonyms) }
        if self.subsets != other.subsets:
            reasons["subsets"] = {"current": value(self.subsets), "previous": value(other.subsets) }
        if self.xrefs != other.xrefs:
            reasons["xrefs"] = {"current": value(self.xrefs), "previous": value(other.xrefs) }
        if self.is_a != other.is_a:
            reasons["is_a"] = {"current": value(self.is_a), "previous": value(other.is_a) }
        if self.relationship != other.relationship:
            reasons["relationship"] = {"current": value(self.relationship), "previous": value(other.relationship) }
        if self.intersection_of != other.intersection_of:
            reasons["intersection_of"] = {"current": value(self.intersection_of), "previous": value(other.intersection_of) }
        return reasons
        
    def __str__(self):
        return self.id + "\t" + self.name
        
class Relation:
    
    def __init__(self):
        id = None
        name = None
        namespace = None
        xref = None
        is_transitive = False
        
    def __str__(self):
        return self.id + "\t" + self.name


class RelationTarget:

    def __init__(self):
        id = None
        relation = None # Relation
        target = None   # NamedEntity


# TODO: I have to add the is_a: term_id ! term_name but I have to add it in the edges of the graph
# TODO: I can also add the consider (who link to other term_ids)
# TODO: Other relations: intersection_of, relationship
class OBO_Parser:
    
    term_key = "[Term]"
    type_def_key = "[Typedef]"
    header = None
    obo_graph = None
    relation_graph = None
    
    def __init__(self, content):
        self.content = content
        self.obo_graph = nx.Graph()
        self.relation_graph = nx.Graph()
        self.header = { }
        self._parseHeader()
        self._parseTerms()
        self._parseRelations()
        print("oboparser: ", len(self.obo_graph) , " terms")
            
            
    def _parseHeader(self):
        lines = self.content.split(self.term_key)[0].split("\n")
        for line in lines:
            if len(line) == 0:
                continue
            kv = re.split(":(?=\s)", line)
#            print("line: " , line , " kv: ", kv)
            self.header[kv[0].strip()] = kv[1].strip()
        print(self.header)


    def _parseTerms(self):
        terms = self.content.split(self.term_key)
        for i in range(1, len(terms)):
            term = Term()

            for line in terms[i].split("\n"):
                if len(line) == 0:
                    continue
                
                if line.startswith(self.type_def_key):
                    break
                
                value = re.split(":(?=\s)", line)[1].strip()
                if line.startswith("id"):
                    term.id = value
                elif line.startswith("alt_id"):
                    term.add_alternate_id(value)
                elif line.startswith("namespace"):
                    term.namespace = value
                elif line.startswith("name"):
                    term.name = value
                elif line.startswith("comment"):
                    term.comment = value
                elif line.startswith("def"):
                    term.definition = value
                elif line.startswith("synonym"):
                    term.add_synonym(value)
                elif line.startswith("subset"):
                    term.add_subset(value)
                elif line.startswith("is_obsolete"):
                    term.is_obsolete = value
                    if term.is_obsolete == "true":
                        term.is_obsolete = True
                elif line.startswith("xref"):
                    term.add_xref(value)
                elif line.startswith("is_a"):
                    term.add_is_a(value.split(" ! ")[0].strip())
                elif line.startswith("relationship"):
                    split = value.split("GO:")
                    split2 = split[1].split(" ! ");
                    target_id = "GO:" + split2[0].strip()
                    target_label = split2[1].strip()
                    term.add_relationship(split[0].strip(), target_id, target_label)
                elif line.startswith("intersection_of"):
                    split = value.split("GO:")
                    split2 = split[1].split(" ! ");
                    target_id = "GO:" + split2[0].strip()
                    target_label = split2[1].strip()
                    term.add_intersection_of(split[0].strip(), target_id, target_label)

            self.obo_graph.add_node(term.id, object=term)
            

    def _parseRelations(self):
        relations = self.content.split(self.type_def_key)

        for i in range(1, len(relations)):
            relation = Relation()

            for line in relations[i].split("\n"):
                if len(line) == 0:
                    continue
                
                value = line.split(":")[1].strip()
                if line.startswith("id"):
                    relation.id = value
                elif line.startswith("name"):
                    relation.name = value
                elif line.startswith("namespace"):
                    relation.namespace = value
                elif line.startswith("xref"):
                    relation.xref = value
                elif line.startswith("is_transitive"):
                    relation.is_transitive = value

            self.relation_graph.add_node(relation.id, object=relation)
            
            
    def get_nodes(self):
        return self.obo_graph.nodes(data=True)


    def get_terms(self, term_state = TermState.VALID):
        map = { }
        for id, data in self.obo_graph.nodes(data=True):
            if term_state == TermState.ANY or (term_state == TermState.OBSOLETED and data['object'].is_obsolete) or (term_state == TermState.VALID and not data['object'].is_obsolete):
                map[id] = data['object']
        return map
        
        
    def get_terms_in(self, aspect, term_state = TermState.VALID):
        list = []
        for id, data in self.obo_graph.nodes(data=True):
            if term_state == TermState.ANY or (term_state == TermState.OBSOLETED and data['object'].is_obsolete) or (term_state == TermState.VALID and not data['object'].is_obsolete):
                if data['object'].namespace == aspect:
                    list.append(data['object'].id)
        return list
        
    def get_merged_terms(self, term_state = TermState.VALID):
        list = set()
        for id, data in self.obo_graph.nodes(data=True):
            if term_state == TermState.ANY or (term_state == TermState.OBSOLETED and data['object'].is_obsolete) or (term_state == TermState.VALID and not data['object'].is_obsolete):
                if data['object'].alt_ids:
                    for alt_id in data['object'].alt_ids:
                        if self.term_used_as_alternate(alt_id):
                            list.add(data['object'].id)
        return list
        
    def has_term(self, query):
        return self.obo_graph.has_node(query)


    def get_term(self, query):
        if not self.has_term(query):
            return None
        return self.obo_graph.node[query]['object']
        
    
    def term_used_as_alternate(self, query):
        for id, data in self.obo_graph.nodes(data=True):
            if data['object'].has_alt_id(query):
                return True
        return False
        
    def get_alternate_terms(self, query):
        list = []
        for id, data in self.obo_graph.nodes(data=True):
            if data['object'].has_alt_id(query):
                list.append({ "id": data['object'].id , "name": data['object'].name })
        return list
        
    def count_all_metas(self, term_state = TermState.VALID, includeXRefs = True):
        count = 0
        for id, data in self.obo_graph.nodes(data=True):
            if term_state == TermState.ANY or (term_state == TermState.OBSOLETED and data['object'].is_obsolete) or (term_state == TermState.VALID and not data['object'].is_obsolete):
                count += data['object'].count_metas(includeXRefs)
        return count
        
        
    def count_all_structurals(self, term_state = TermState.VALID):
        count = 0
        for id, data in self.obo_graph.nodes(data=True):
            if term_state == TermState.ANY or (term_state == TermState.OBSOLETED and data['object'].is_obsolete) or (term_state == TermState.VALID and not data['object'].is_obsolete):
                count += data['object'].count_structurals()
        return count

    def get_children(self, root):
        children = set()
        for id, data in self.obo_graph.nodes(data=True):
            current = data['object']
            if current.is_obsolete:
                continue
            if current.is_a is None:
                continue
            if root.id in current.is_a:
                children.add(current)
                # locals = self.get_children(current)
                # for local in locals:
                #     children.add(local)
        return children


def main(argv):
    go_obo_url = "https://geneontology-public.s3.amazonaws.com/archive/2019-06-09_go.obo"
    req = requests.get(go_obo_url)
    print("obo downloaded")

    obo = OBO_Parser(req.text)
    term_id = "GO:0005515"
    term = obo.get_term(term_id)

    print("using term: ", term.id , term.name)
    derived_terms = obo.get_children(term)
    print(len(derived_terms) , " derived terms")

    derived_ids = []
    for dterm in derived_terms:
        derived_ids.append(dterm.id)

    not_derived_terms = " AND NOT \"" + "\" AND NOT \"".join(derived_ids)


    url = "http://golr-aux.geneontology.io/solr/" + "select?fq=document_category:\"bioentity\"&q=*:*&wt=json&facet=true&facet.field=type&facet.field=taxon&facet.limit=1000000&facet.mincount=1&rows=10&fq=!annotation_class_list:(\"GO:0005515\"" + not_derived_terms + "\")"
    print(url)

    req = requests.get(url)
    print(req.text)


if __name__ == "__main__":
   main(sys.argv[1:])




# GO Store Changes
# sudo yum install python34-setuptools
# sudo yum install python34-devel
# pip3 install pyyaml -t .
# sudo python3 -m pip install pyyaml

from obo_parser import OBO_Parser, TermState
import sys, getopt, os

import requests
import json

# import boto3
# import botocore
# go_s3_bucket_name = "geneontology"
# go_obo_key = "go.obo"
# s3_resource = boto3.resource('s3')
# go_s3_bucket = s3_resource.Bucket(name=go_s3_bucket_name)
# from gzip import GzipFile
# from io import BytesIO


#go_doi_url = "https://zenodo.org/api/records/1205166"

go_pipeline_release_url = "http://current.geneontology.org/metadata/release-date.json"
go_obo_url = "http://purl.obolibrary.org/obo/go.obo"

release_date = "N/A"
last_obo = None
last_date = None

# def lambda_handlerLIST(event, context):
#     global release_date
#     release_date = get_release_date()
    
#     global last_obo
#     last_obo = get_last_obo()


#     static_list = [
#         "2018-07-02",
#         "2018-08-09",
#         "2018-09-05",
#         "2018-10-08",
#         "2018-11-15",
#         "2018-12-01",
#         "2019-01-01",
#         "2019-02-01",
#         "2019-03-18",
#         "2019-04-17",
#         "2019-05-09",
#         "2019-06-09"
#     ]
    
#     global last_date
    
#     for i in range(1, len(static_list)):
#         print(i , static_list[i] , " vs " , static_list[i-1])
#         last_date = static_list[i-1]
#         old_obo_url = "https://s3.amazonaws.com/" + go_s3_bucket_name +"/archive/" + last_date + "_go.obo"
        
#         release_date = static_list[i]
#         new_obo_url = "https://s3.amazonaws.com/" + go_s3_bucket_name +"/archive/" + release_date + "_go.obo"

#         try:
#             save_changes(new_obo_url, old_obo_url)
#         except:
#             print("ERROR while generating diff (" + new_obo_url + ", " + old_obo_url + ")")
    
#     return 'success'
    
# def lambda_handler(event, context):
#     global release_date
#     release_date = get_release_date()
    
#     global last_obo
#     last_obo = get_last_obo()

#     new_obo_url = "http://purl.obolibrary.org/obo/go.obo"
#     old_obo_url = "https://s3.amazonaws.com/" + go_s3_bucket_name + "/archive/" + last_obo

#     store_current_go_obo()
#     save_changes(new_obo_url, old_obo_url)
#     return 'success'    
    
def compute_changes(current_obo_url, previous_obo_url):
    # The new published OBO archive
    print("Loading current GO ontology (" + current_obo_url + ")...")
    currentgo = OBO_Parser(requests.get(current_obo_url).text)

    # A previously published OBO archive
    print("Loading previous GO ontology (" + previous_obo_url + ")...")
    oldgo = OBO_Parser(requests.get(previous_obo_url).text)

    # New GO Terms
    created = { }
    created_count = 0
    for id, newterm in currentgo.get_terms().items():                                                                                                                                                                                
        if not oldgo.has_term(id):
            if newterm.namespace not in created:
                created[newterm.namespace] = []
            created[newterm.namespace].append({ "id": id, "name": newterm.name})
            created_count += 1
    print(str(created_count) + " terms created since last revision")

    # Merged GO Terms    
    merged = { }
    merged_list = []
    merged_count = 0
    for id, oldterm in oldgo.get_terms().items():                                                                                                                                                                                  
        if not currentgo.has_term(id):
            if oldterm.namespace not in merged:
                merged[oldterm.namespace] = []
            alts = currentgo.get_alternate_terms(id)
            if len(alts) > 0:
                merged[oldterm.namespace].append( { "current": alts[0], "previous": { "id": id, "name": oldterm.name } } )
                merged_count += 1
                merged_list.append(oldterm.id)
    print(str(merged_count) + " terms merged since last revision")
    
    # Obsoleted GO Terms
    obsoleted = { }
    obsoleted_count = 0
    new_terms = currentgo.get_terms()
    for id, oldterm in oldgo.get_terms().items():                                                                                                                                                                                  
        if id not in new_terms:
            if oldterm.namespace not in obsoleted:
                obsoleted[oldterm.namespace] = []
            if oldterm.id not in merged_list:
                obsoleted[oldterm.namespace].append({ "id": id, "name": oldterm.name})
                obsoleted_count += 1
    print(str(obsoleted_count) + " terms obsoleted since last revision")
    
    # Existing GO Terms with structural changes (is_a, part_of, has_part etc)
    relations_changes = { }
    structural_count = 0
    structural_total_count = 0
    for id, newterm in currentgo.get_terms().items():
        if oldgo.has_term(id):                                                                                                                                                                            
            oldterm = oldgo.get_term(id)
            if not newterm.structural_equals(oldterm):
                if newterm.namespace not in relations_changes:
                    relations_changes[newterm.namespace] = []
                    
                reasons = {}
                for key, reason in newterm.explain_structural_differences(oldterm).items():
                    reasons[key] = { "current" : reason['current'], "previous" : reason['previous'] }
                relations_changes[newterm.namespace].append({ "id" : id, "name": newterm.name , "changes": reasons })
                structural_count += 1
                structural_total_count += len(reasons)
    print(str(structural_count) + " terms relation changes since last revision")
    
    
    # Existing GO Terms with meta changes (synonyms, xrefs, definition, etc)
    meta_changes = { }
    meta_count = 0
    meta_total_count = 0
    for id, newterm in currentgo.get_terms().items():
        if oldgo.has_term(id):                                                                                                                                                                            
            oldterm = oldgo.get_term(id)
            if not newterm.meta_equals(oldterm):
                if newterm.namespace not in meta_changes:
                    meta_changes[newterm.namespace] = []
                    
                reasons = {}
                for key, reason in newterm.explain_meta_differences(oldterm).items():
                    reasons[key] = { "current" : reason['current'], "previous" : reason['previous'] }
                meta_changes[newterm.namespace].append({ "id" : id, "name": newterm.name , "changes": reasons })
                meta_count += 1
                meta_total_count += len(reasons)
    print(str(meta_count) + " terms meta changes since last revision")

    # Existing GO Terms with meta changes (synonyms, NO XREFS, definition, etc)
    meta_noxrefs_changes = { }
    meta_noxrefs_count = 0
    meta_noxrefs_total_count = 0
    for id, newterm in currentgo.get_terms().items():
        if oldgo.has_term(id):                                                                                                                                                                            
            oldterm = oldgo.get_term(id)
            if not newterm.meta_equals(oldterm, False):
                if newterm.namespace not in meta_noxrefs_changes:
                    meta_noxrefs_changes[newterm.namespace] = []
                    
                reasons = {}
                for key, reason in newterm.explain_meta_differences(oldterm, False).items():
                    reasons[key] = { "current" : reason['current'], "previous" : reason['previous'] }
                meta_noxrefs_changes[newterm.namespace].append({ "id" : id, "name": newterm.name , "changes": reasons })
                meta_noxrefs_count += 1                
                meta_noxrefs_total_count += len(reasons)
    print(str(meta_noxrefs_count) + " terms meta (NO XREFS) changes since last revision")
 

    release_date = currentgo.header['data-version']
    release_date = release_date[release_date.index("/") + 1:] if "/" in release_date else release_date

    last_date = oldgo.header['data-version']
    last_date = last_date[last_date.index("/") + 1:] if "/" in last_date else last_date

    print("Creating JSON report...")
    report = { }
    report["releases_compared"] = {
                "current" : { "date": release_date, "version" : currentgo.header['data-version'], "format" : currentgo.header['format-version'] },
                "previous" : { "date" : last_date, "version" : oldgo.header['data-version'], "format" : oldgo.header['format-version'] }
            }
    report["summary"] = {
        "current": {
            "release_date" : release_date,
            "valid_terms" : len(currentgo.get_terms(TermState.VALID)),
            "obsoleted_terms" : len(currentgo.get_terms(TermState.OBSOLETED)),
            "merged_terms" : len(currentgo.get_merged_terms(TermState.ANY)),
            "biological_process_terms" : len(currentgo.get_terms_in("biological_process")),
            "molecular_function_terms" : len(currentgo.get_terms_in("molecular_function")),
            "cellular_component_terms" : len(currentgo.get_terms_in("cellular_component")),
            "meta_statements" : currentgo.count_all_metas(),
            "meta_statements_exclude_xrefs" : currentgo.count_all_metas(TermState.VALID, False),
            "terms_relations" : currentgo.count_all_structurals()
        },
        "previous": {
            "release_date" : last_date,
            "valid_terms" : len(oldgo.get_terms(TermState.VALID)),
            "obsoleted_terms" : len(oldgo.get_terms(TermState.OBSOLETED)),
            "merged_terms" : len(oldgo.get_merged_terms(TermState.ANY)),
            "biological_process_terms" : len(oldgo.get_terms_in("biological_process")),
            "molecular_function_terms" : len(oldgo.get_terms_in("molecular_function")),
            "cellular_component_terms" : len(oldgo.get_terms_in("cellular_component")),
            "meta_statements" : oldgo.count_all_metas(),
            "meta_statements_exclude_xrefs" : oldgo.count_all_metas(TermState.VALID, False),
            "terms_relations" : oldgo.count_all_structurals()
        },
        "changes" : {
            "created_terms" : created_count,
            "obsoleted_terms" : obsoleted_count,
            "merged_terms" : merged_count,
            "relations_changes_by_term" : structural_count,
            "meta_changes_by_term" : meta_count,
            "meta_changes_exclude_xrefs_by_term" : meta_noxrefs_count,
            "relations_changes" : structural_total_count,
            "meta_changes" : meta_total_count,
            "meta_changes_exclude_xrefs" : meta_noxrefs_total_count
        }
    }
    report["detailed_changes"] = {
        "created_terms" : created,
        "obsoleted_terms" : obsoleted,
        "merged_terms" : merged,
        "relations_changes" : relations_changes,
        "meta_changes" : meta_changes,
        "meta_changes_exclude_xrefs" : meta_noxrefs_changes
    }

    print("JSON report created.")

    return report


def flattern(A):
    rt = []
    for i in A:
        if isinstance(i,list): rt.extend(flattern(i))
        else: rt.append(i)
    return rt

# def save_changes(new_obo_url, old_obo_url):

#     # The new published OBO archive
#     print("Loading current GO ontology (" + new_obo_url + ")...")
#     newgo = OBO_Parser(requests.get(new_obo_url).text)

#     # A previously published OBO archive
#     print("Loading previous GO ontology (" + old_obo_url + ")...")
#     oldgo = OBO_Parser(requests.get(old_obo_url).text)
    
    
#     # New GO Terms
#     created = { }
#     created_count = 0
#     for id, newterm in newgo.get_terms().items():                                                                                                                                                                                
#         if not oldgo.has_term(id):
#             if newterm.namespace not in created:
#                 created[newterm.namespace] = []
#             created[newterm.namespace].append({ "id": id, "name": newterm.name})
#             created_count += 1
#     print(str(created_count) + " terms created since last revision")

#     # Merged GO Terms    
#     merged = { }
#     merged_list = []
#     merged_count = 0
#     for id, oldterm in oldgo.get_terms().items():                                                                                                                                                                                  
#         if not newgo.has_term(id):
#             if oldterm.namespace not in merged:
#                 merged[oldterm.namespace] = []
#             alts = newgo.get_alternate_terms(id)
#             if len(alts) > 0:
#                 merged[oldterm.namespace].append( { "current": alts[0], "previous": { "id": id, "name": oldterm.name } } )
#                 merged_count += 1
#                 merged_list.append(oldterm.id)
#     print(str(merged_count) + " terms merged since last revision")
    
#     # Obsoleted GO Terms
#     obsoleted = { }
#     obsoleted_count = 0
#     new_terms = newgo.get_terms()
#     for id, oldterm in oldgo.get_terms().items():                                                                                                                                                                                  
#         if id not in new_terms:
#             if oldterm.namespace not in obsoleted:
#                 obsoleted[oldterm.namespace] = []
#             if oldterm.id not in merged_list:
#                 obsoleted[oldterm.namespace].append({ "id": id, "name": oldterm.name})
#                 obsoleted_count += 1
#     print(str(obsoleted_count) + " terms obsoleted since last revision")
    
#     # Existing GO Terms with structural changes (is_a, part_of, has_part etc)
#     relations_changes = { }
#     structural_count = 0
#     structural_total_count = 0
#     for id, newterm in newgo.get_terms().items():
#         if oldgo.has_term(id):                                                                                                                                                                            
#             oldterm = oldgo.get_term(id)
#             if not newterm.structural_equals(oldterm):
#                 if newterm.namespace not in relations_changes:
#                     relations_changes[newterm.namespace] = []
                    
#                 reasons = {}
#                 for key, reason in newterm.explain_structural_differences(oldterm).items():
#                     reasons[key] = { "current" : reason['current'], "previous" : reason['previous'] }
#                 relations_changes[newterm.namespace].append({ "id" : id, "name": newterm.name , "changes": reasons })
#                 structural_count += 1
#                 structural_total_count += len(reasons)
#     print(str(structural_count) + " terms structural changes since last revision")
    
    
#     # Existing GO Terms with meta changes (synonyms, xrefs, definition, etc)
#     meta_changes = { }
#     meta_count = 0
#     meta_total_count = 0
#     for id, newterm in newgo.get_terms().items():
#         if oldgo.has_term(id):                                                                                                                                                                            
#             oldterm = oldgo.get_term(id)
#             if not newterm.meta_equals(oldterm):
#                 if newterm.namespace not in meta_changes:
#                     meta_changes[newterm.namespace] = []
                    
#                 reasons = {}
#                 for key, reason in newterm.explain_meta_differences(oldterm).items():
#                     reasons[key] = { "current" : reason['current'], "previous" : reason['previous'] }
#                 meta_changes[newterm.namespace].append({ "id" : id, "name": newterm.name , "changes": reasons })
#                 meta_count += 1
#                 meta_total_count += len(reasons)
#     print(str(meta_count) + " terms meta changes since last revision")

#     # Existing GO Terms with meta changes (synonyms, NO XREFS, definition, etc)
#     meta_noxrefs_changes = { }
#     meta_noxrefs_count = 0
#     meta_noxrefs_total_count = 0
#     for id, newterm in newgo.get_terms().items():
#         if oldgo.has_term(id):                                                                                                                                                                            
#             oldterm = oldgo.get_term(id)
#             if not newterm.meta_equals(oldterm, False):
#                 if newterm.namespace not in meta_noxrefs_changes:
#                     meta_noxrefs_changes[newterm.namespace] = []
                    
#                 reasons = {}
#                 for key, reason in newterm.explain_meta_differences(oldterm, False).items():
#                     reasons[key] = { "current" : reason['current'], "previous" : reason['previous'] }
#                 meta_noxrefs_changes[newterm.namespace].append({ "id" : id, "name": newterm.name , "changes": reasons })
#                 meta_noxrefs_count += 1                
#                 meta_noxrefs_total_count += len(reasons)
#     print(str(meta_noxrefs_count) + " terms meta (NO XREFS) changes since last revision")
 
#     report = { }
#     report["releases"] = {
#                 "current" : { "date": release_date, "version" : newgo.header['data-version'], "format" : newgo.header['format-version'] },
#                 "previous" : { "date" : last_date, "version" : oldgo.header['data-version'], "format" : oldgo.header['format-version'] }
#             }
#     report["summary"] = {
#         "current": {
#             "release_date" : report["releases"]["current"]["version"],
#             "valid_terms" : len(newgo.get_terms(TermState.VALID)),
#             "obsoleted_terms" : len(newgo.get_terms(TermState.OBSOLETED)),
#             "merged_terms" : len(newgo.get_merged_terms(TermState.ANY)),
#             "biological_processes" : len(newgo.get_terms_in("biological_process")),
#             "molecular_functions" : len(newgo.get_terms_in("molecular_function")),
#             "cellular_components" : len(newgo.get_terms_in("cellular_component")),
#             "meta_statements" : newgo.count_all_metas(),
#             "meta_noxrefs_statements" : newgo.count_all_metas(TermState.VALID, False),
#             "structural_statements" : newgo.count_all_structurals()
#         },
#         "previous": {
#             "release_date" : report["releases"]["previous"]["version"],
#             "valid_terms" : len(oldgo.get_terms(TermState.VALID)),
#             "obsoleted_terms" : len(oldgo.get_terms(TermState.OBSOLETED)),
#             "merged_terms" : len(oldgo.get_merged_terms(TermState.ANY)),
#             "biological_processes" : len(oldgo.get_terms_in("biological_process")),
#             "molecular_functions" : len(oldgo.get_terms_in("molecular_function")),
#             "cellular_components" : len(oldgo.get_terms_in("cellular_component")),
#             "meta_statements" : oldgo.count_all_metas(),
#             "meta_noxrefs_statements" : oldgo.count_all_metas(TermState.VALID, False),
#             "structural_statements" : oldgo.count_all_structurals()
#         },
#         "created_terms" : created_count,
#         "obsoleted_terms" : obsoleted_count,
#         "merged_terms" : merged_count,
#         "relations_changes_by_term" : structural_count,
#         "meta_changes_by_term" : meta_count,
#         "terms_meta_noxrefs_changes" : meta_noxrefs_count,
#         "relations_changes" : structural_total_count,
#         "meta_changes" : meta_total_count,
#         "total_meta_noxrefs_changes" : meta_noxrefs_total_count
#     }
#     report["created_terms"] = created
#     report["obsoleted_terms"] = obsoleted
#     report["merged_terms"] = merged
#     report["relations_changes"] = relations_changes
#     report["meta_changes"] = meta_changes
#     report["meta_noxrefs_changes"] = meta_noxrefs_changes
    
#     # TO READD    
#     store_json("go-last-changes.json", report)
#     store_json("go-last-changes-summary.json", report["summary"])
#     store_json("archive/" + release_date + "_go-last-changes.json", report)
#     store_json("archive/" + release_date + "_go-last-changes-summary.json", report["summary"])

    

 
#     # Creating TSV version of the JSON report
#     txtreport = "CHANGES IN GO ONTOLOGY BETWEEN\ncurrent\t" + report["releases"]["current"]["version"] + " (" + release_date + ")\nprevious\t" + report["releases"]["previous"]["version"] + " (" + last_date + ")\n"

#     txtreport += "\nSUMMARY\n"
#     txtreport += "current_valid_terms\t" + str(report["summary"]["current"]["valid_terms"]) + "\n"
#     txtreport += "current_obsoleted_terms\t" + str(report["summary"]["current"]["obsoleted_terms"]) + "\n"
#     txtreport += "current_merged_terms\t" + str(report["summary"]["current"]["merged_terms"]) + "\n"
#     txtreport += "current_biological_processes\t" + str(report["summary"]["current"]["biological_processes"]) + "\n"
#     txtreport += "current_molecular_functions\t" + str(report["summary"]["current"]["molecular_functions"]) + "\n"
#     txtreport += "current_cellular_components\t" + str(report["summary"]["current"]["cellular_components"]) + "\n"
#     txtreport += "current_meta_statements\t" + str(report["summary"]["current"]["meta_statements"]) + "\n"
#     txtreport += "current_meta_noxrefs_statements\t" + str(report["summary"]["current"]["meta_noxrefs_statements"]) + "\n"
#     txtreport += "current_structural_statements\t" + str(report["summary"]["current"]["structural_statements"]) + "\n"

#     txtreport += "previous_valid_terms\t" + str(report["summary"]["previous"]["valid_terms"]) + "\n"
#     txtreport += "previous_obsoleted_terms\t" + str(report["summary"]["previous"]["obsoleted_terms"]) + "\n"
#     txtreport += "previous_merged_terms\t" + str(report["summary"]["previous"]["merged_terms"]) + "\n"
#     txtreport += "previous_biological_processes\t" + str(report["summary"]["previous"]["biological_processes"]) + "\n"
#     txtreport += "previous_molecular_functions\t" + str(report["summary"]["previous"]["molecular_functions"]) + "\n"
#     txtreport += "previous_cellular_components\t" + str(report["summary"]["previous"]["cellular_components"]) + "\n"
#     txtreport += "previous_meta_statements\t" + str(report["summary"]["previous"]["meta_statements"]) + "\n"
#     txtreport += "previous_meta_noxrefs_statements\t" + str(report["summary"]["previous"]["meta_noxrefs_statements"]) + "\n"
#     txtreport += "previous_structural_statements\t" + str(report["summary"]["previous"]["structural_statements"]) + "\n"

#     txtreport += "created_terms\t" + str(report["summary"]["created_terms"]) + "\n"
#     txtreport += "obsoleted_terms\t" + str(report["summary"]["obsoleted_terms"]) + "\n"
#     txtreport += "merged_terms\t" + str(report["summary"]["merged_terms"]) + "\n"
#     txtreport += "relations_changes_by_term\t" + str(report["summary"]["relations_changes_by_term"]) + "\n"
#     txtreport += "meta_changes_by_term\t" + str(report["summary"]["meta_changes_by_term"]) + "\n"
#     txtreport += "terms_meta_noxrefs_changes\t" + str(report["summary"]["terms_meta_noxrefs_changes"]) + "\n"
#     txtreport += "relations_changes\t" + str(report["summary"]["relations_changes"]) + "\n"
#     txtreport += "meta_changes\t" + str(report["summary"]["meta_changes"]) + "\n"
#     txtreport += "total_meta_noxrefs_changes\t" + str(report["summary"]["total_meta_noxrefs_changes"]) + "\n"

#     txtreport += "\nTERMS CREATED\t" + str(created_count) + "\n"
#     for aspect in report["created_terms"]:
#         for term in report["created_terms"][aspect]:
#             txtreport += aspect + "\t" + term["id"] + "\t" + term["name"] + "\n"
                    
#     txtreport += "\nTERMS OBSOLETED\t" + str(obsoleted_count) + "\n"
#     for aspect in report["obsoleted_terms"]:
#         for term in report["obsoleted_terms"][aspect]:
#             txtreport += aspect + "\t" + term["id"] + "\t" + term["name"] + "\n"

#     txtreport += "\nTERMS MERGED\t" + str(merged_count) + "\n"
#     for aspect in report["merged_terms"]:
#         for term in report["merged_terms"][aspect]:
#             txtreport += aspect + "\t" + term["current"]["id"] + "\t" + term["current"]["name"] + "\t" + term["previous"]["id"] + "\t" + term["previous"]["name"] + "\n"

#     txtreport += "\nTERMS STRUCTURAL CHANGES\t" + str(structural_count) + "\n"
#     for aspect in report["relations_changes"]:
#         for term in report["relations_changes"][aspect]:
#             txtreport += aspect + "\t" + term["id"] + "\t" + term["name"]
#             for change in term["changes"]:
#                 curr = term["changes"][change]["current"]
#                 if not isinstance(curr, str):
#                     curr = ", ".join(map(str, curr))
                
#                 prev = term["changes"][change]["previous"]
#                 if not isinstance(prev, str):
#                     prev = ", ".join(map(str, prev))
#                 txtreport += "\t" + change + " (current: " + curr + " previous: " + prev + ")"
#             txtreport += "\n"
        
#     txtreport += "\nTERMS META CHANGES\t" + str(meta_count) + "\n"
#     for aspect in report["meta_changes"]:
#         for term in report["meta_changes"][aspect]:
#             txtreport += aspect + "\t" + term["id"] + "\t" + term["name"]
#             for change in term["changes"]:
#                 curr = term["changes"][change]["current"]
#                 if not isinstance(curr, str):
#                     curr = ", ".join(map(str, curr))
                
#                 prev = term["changes"][change]["previous"]
#                 if not isinstance(prev, str):
#                     prev = ", ".join(map(str, prev))
#                 txtreport += "\t" + change + " (current: " + curr + " previous: " + prev + ")"
#             txtreport += "\n"

#     txtreport += "\nTERMS META CHANGES (NO XREFS)\t" + str(meta_noxrefs_count) + "\n"
#     for aspect in report["meta_noxrefs_changes"]:
#         for term in report["meta_noxrefs_changes"][aspect]:
#             txtreport += aspect + "\t" + term["id"] + "\t" + term["name"]
#             for change in term["changes"]:
#                 curr = term["changes"][change]["current"]
#                 if not isinstance(curr, str):
#                     curr = ", ".join(map(str, curr))
                
#                 prev = term["changes"][change]["previous"]
#                 if not isinstance(prev, str):
#                     prev = ", ".join(map(str, prev))
#                 txtreport += "\t" + change + " (current: " + curr + " previous: " + prev + ")"
#             txtreport += "\n"
            
#     store_text("go-last-changes.tsv", txtreport)
#     store_text("archive/" + release_date + "_go-last-changes.tsv", txtreport)



def create_text_report(json_changes):
    text_report = ""

    text_report = "CHANGES IN GO ONTOLOGY"
    
    text_report += "\n\nRELEASES COMPARED"
    text_report += "\ncurrent_release_date\t" + json_changes["releases_compared"]["current"]["date"]    
    text_report += "\nprevious_release_date\t" + json_changes["releases_compared"]["previous"]["date"]

    text_report += "\n\nSUMMARY: CURRENT RELEASE"
    for key, val in json_changes["summary"]["current"].items():
        if "release_date" != key:
            text_report += "\n" + key + "\t" + str(val)

    text_report += "\n\nSUMMARY: PREVIOUS RELEASE"
    for key, val in json_changes["summary"]["previous"].items():
        if "release_date" != key:
            text_report += "\n" + key + "\t" + str(val)

    text_report += "\n\nSUMMARY: DIFF BETWEEN RELEASES"
    for key, val in json_changes["summary"]["changes"].items():
        text_report += "\n" + key + "\t" + str(val)

    text_report += "\n\nDETAILED CHANGES"

    text_report += "\n\n" + count(json_changes["detailed_changes"]["created_terms"]) + " CREATED TERMS"
    for key, val in json_changes["detailed_changes"]["created_terms"].items():
        for item in val:
            text_report += "\n" + key + "\t" + item["id"] + "\t" + item["name"]

    text_report += "\n\n" + count(json_changes["detailed_changes"]["obsoleted_terms"]) + " OBSOLETED TERMS"
    for key, val in json_changes["detailed_changes"]["obsoleted_terms"].items():
        for item in val:
            text_report += "\n" + key + "\t" + item["id"] + "\t" + item["name"]

    text_report += "\n\n" + count(json_changes["detailed_changes"]["merged_terms"]) + " MERGED TERMS"
    for key, val in json_changes["detailed_changes"]["merged_terms"].items():
        for item in val:
            text_report += "\n" + key + "\t" + item["current"]["id"] + "\t" + item["current"]["name"] + "\tWAS\t" + "\t" + item["previous"]["id"] + "\t" + item["previous"]["name"]

    text_report += "\n\n" + count(json_changes["detailed_changes"]["relations_changes"]) + " RELATION CHANGES"
    for key, val in json_changes["detailed_changes"]["relations_changes"].items():
        for item in val:
            text_report += "\n" + key + "\t" + item["id"] + "\t" + item["name"]
            data = item["changes"]
            for field in data:
                text_report += "\n\t" + field + "\t" + format(data[field]["current"]) + "\tWAS\t" +format(data[field]["previous"])

    text_report += "\n\n" + count(json_changes["detailed_changes"]["meta_changes_exclude_xrefs"]) + " META CHANGES EXCLUDING XREFS"
    for key, val in json_changes["detailed_changes"]["meta_changes_exclude_xrefs"].items():
        for item in val:
            text_report += "\n" + key + "\t" + item["id"] + "\t" + item["name"]
            data = item["changes"]
            for field in data:
                text_report += "\n\t" + field + "\t" + format(data[field]["current"]) + "\tWAS\t" +format(data[field]["previous"])

    text_report += "\n\n" + count(json_changes["detailed_changes"]["meta_changes"]) + " META CHANGES"
    for key, val in json_changes["detailed_changes"]["meta_changes"].items():
        for item in val:
            text_report += "\n" + key + "\t" + item["id"] + "\t" + item["name"]
            data = item["changes"]
            for field in data:
                text_report += "\n\t" + field + "\t" + format(data[field]["current"]) + "\tWAS\t" +format(data[field]["previous"])

    return text_report

def format(item):
    if type(item) == str:
        return item.strip()
    if type(item) == list:
        return ";".join(item)
    return item

def count(map):
    count = 0
    for key in map:
        count += len(map[key])
    return str(count)



def write_json(key, content):
    with open(key, 'w') as outfile:
        json.dump(content, outfile, indent=2)
 
def write_text(key, content):
    with open(key, 'w') as outfile:
        outfile.write(content)



def print_help():
    print('\nUsage: python go_ontology_changes.py -c <current_obo_url> -p <previous_obo_url> -o <output_rep>\n')


def main(argv):
    current_obo_url = ''
    previous_obo_url = ''
    output_rep = ''

    if len(argv) < 6:
        print_help()
        sys.exit(2)

    try:
        opts, argv = getopt.getopt(argv,"c:p:o:",["cobo=","pobo=","orep="])
    except getopt.GetoptError:
        print_help()
        sys.exit(2)

    for opt, arg in opts:
        if opt == '-h':
            print_help()
            sys.exit()
        elif opt in ("-c", "--cobo"):
            current_obo_url = arg
        elif opt in ("-p", "--pobo"):
            previous_obo_url = arg
        elif opt in ("-o", "-orep"):
            output_rep = arg
        
    if not output_rep.endswith("/"):
        output_rep += "/"

    if not os.path.exists(output_rep):
        os.mkdir(output_rep)

    output_json =  output_rep + "go-ontology-changes.json"
    output_stats_json = output_rep + "go-ontology-stats.json"
    output_tsv =  output_rep + "go-ontology-changes.tsv"

    print("Will write ontology changes to " + output_json + " and " + output_tsv)

    json_changes = compute_changes(current_obo_url, previous_obo_url)

    print("Saving Stats to <" + output_json + "> ...")    
    write_json(output_json, json_changes)
    write_json(output_stats_json, json_changes["summary"]["current"])
    print("Done.")

    print("Saving Stats to <" + output_tsv + "> ...")    
    tsv_changes = create_text_report(json_changes)
    write_text(output_tsv, tsv_changes)
    print("Done.")
    


if __name__ == "__main__":
   main(sys.argv[1:])


import requests
import json
import sys, getopt, os

import go_stats
import go_ontology_changes
import go_annotation_changes


def merge_dict(dict_total, dict_diff):
    new_dict = { }
    for key, val in dict_total.items():
        if type(val) == str:
            new_dict[key] = val
        elif type(val) == int or type(val) == float:
            if val == 0:
                diff_val = dict_diff[key] if key in dict_diff else 0
                new_dict[key] = str(diff_val) + " / " + str(val) + " (0%) "
            else:
                diff_val = dict_diff[key] if key in dict_diff else 0
                new_dict[key] = str(diff_val) + " / " + str(val) + " (" + str(round(100 * diff_val / val, 2)) + "%) "
        elif type(val) == dict:
            diff_val = dict_diff[key] if key in dict_diff else { }
            new_dict[key] = merge_dict(val, diff_val)
        else:
            print("should not happened ! " , val , type(val))
    return new_dict

def minus_dict(dict1, dict2):
    new_dict = { }
    for key, val in dict1.items():
        if type(val) == str:
            new_dict[key] = val
        elif type(val) == int or type(val) == float:
                diff_val = dict2[key] if key in dict2 else 0
                new_dict[key] = val - diff_val
        elif type(val) == dict:
            diff_val = dict2[key] if key in dict2 else { }
            new_dict[key] = merge_dict(val, diff_val)
        else:
            print("should not happened ! " , val , type(val))
    return new_dict    


def alter_annotation_changes(current_stats, previous_stats, json_annot_changes):
    altered_json_annot_changes = {
        "releases_compared" : {
            "current" : current_stats["release_date"],
            "previous" : previous_stats["release_date"]
        },
        "summary" : {
            "current" : {
                "annotations" : {
                    "total" : current_stats["annotations"]["total"],
                    "by_aspect" : current_stats["annotations"]["by_aspect"],
                    "by_evidence_cluster" : current_stats["annotations"]["by_evidence"]["cluster"],
                },
                "bioentities" : current_stats["annotations"]["bioentities"]["total"],
                "taxa" : current_stats["annotations"]["taxa"]["total"],
                "taxa_filtered" : current_stats["annotations"]["taxa"]["filtered"],
                "references" : current_stats["annotations"]["references"]["all"]["total"],
                "pmids" : current_stats["annotations"]["references"]["pmids"]["total"]
            },
            "previous" : {
                "annotations" : {
                    "total" : previous_stats["annotations"]["total"],
                    "by_aspect" : previous_stats["annotations"]["by_aspect"],
                    "by_evidence_cluster" : previous_stats["annotations"]["by_evidence"]["cluster"],
                },
                "bioentities" : previous_stats["annotations"]["bioentities"]["total"],
                "taxa" : previous_stats["annotations"]["taxa"]["total"],
                "taxa_filtered" : previous_stats["annotations"]["taxa"]["filtered"],
                "references" : previous_stats["annotations"]["references"]["all"]["total"],
                "pmids" : previous_stats["annotations"]["references"]["pmids"]["total"]
            },
            "changes" : {
                "annotations" : {
                    "total" : current_stats["annotations"]["total"] - previous_stats["annotations"]["total"],
                    "by_aspect" : minus_dict(current_stats["annotations"]["by_aspect"], previous_stats["annotations"]["by_aspect"]),
                    "by_evidence_cluster" : minus_dict(current_stats["annotations"]["by_evidence"]["cluster"], previous_stats["annotations"]["by_evidence"]["cluster"]),
                },
                "bioentities" : current_stats["annotations"]["bioentities"]["total"] - previous_stats["annotations"]["bioentities"]["total"],
                "taxa" : current_stats["annotations"]["taxa"]["total"] - previous_stats["annotations"]["taxa"]["total"],
                "taxa_filtered" : current_stats["annotations"]["taxa"]["filtered"] - previous_stats["annotations"]["taxa"]["filtered"],
                "references" : current_stats["annotations"]["references"]["all"]["total"] - previous_stats["annotations"]["references"]["all"]["total"],
                "pmids" : current_stats["annotations"]["references"]["pmids"]["total"] - previous_stats["annotations"]["references"]["pmids"]["total"]
            },
        },
        "detailed_changes" : json_annot_changes["annotations"] 
    }
    return altered_json_annot_changes  


def print_help():
    print('\nUsage: python go_reports.py -g <golr_url> -d <release_date> -s <previous_stats_url> -n <previous_stats_no_pb_url> -c <current_obo_url> -p <previous_obo_url> -o <output_rep>\n')


def main(argv):
    golr_url = ''
    previous_stats_url = ''
    previous_stats_no_pb_url = ''
    current_obo_url = ''
    previous_obo_url = ''    
    output_rep = ''
    release_date = ''

    print(len(argv))
    if len(argv) < 14:
        print_help()
        sys.exit(2)

    try:
        opts, argv = getopt.getopt(argv,"g:s:n:c:p:o:d:",["golrurl=", "pstats=", "pnstats=", "cobo=", "pobo=", "orep=", "date="])
    except getopt.GetoptError:
        print_help()
        sys.exit(2)

    for opt, arg in opts:
        if opt == '-h':
            print_help()
            sys.exit()
        elif opt in ("-g", "--golrurl"):
            golr_url = arg
        elif opt in ("-s", "--pstats"):
            previous_stats_url = arg
        elif opt in ("-n", "--pnstats"):
            previous_stats_no_pb_url = arg
        elif opt in ("-c", "--cobo"):
            current_obo_url = arg
        elif opt in ("-p", "--pobo"):
            previous_obo_url = arg
        elif opt in ("-o", "--orep"):
            output_rep = arg
        elif opt in ("-d", "--date"):
            release_date = arg

    if not output_rep.endswith("/"):
        output_rep += "/"

    if not os.path.exists(output_rep):
        os.mkdir(output_rep)


    # 1 - Executing go_stats script
    print("\n\n1a - EXECUTING GO_STATS SCRIPT (INCLUDING PROTEIN BINDING)...\n")
    json_stats = go_stats.compute_stats(golr_url, release_date)
    # with open('newstats/2019-06/go-stats.json', 'r') as myfile:
    #     data=myfile.read()
    # json_stats = json.loads(data)
    
    print("DONE.")

    print("\n\n1b - EXECUTING GO_STATS SCRIPT (EXCLUDING PROTEIN BINDING)...\n")
    json_stats_no_pb = go_stats.compute_stats(golr_url, release_date, True)
    # with open('newstats/2019-06/go-stats-no-pb.json', 'r') as myfile:
    #     data=myfile.read()
    # json_stats_no_pb = json.loads(data)
    # print("DONE.")


    # 2 - Executing go_ontology_changes script
    print("\n\n2 - EXECUTING GO_ONTOLOGY_CHANGES SCRIPT...\n")
    json_onto_changes = go_ontology_changes.compute_changes(current_obo_url, previous_obo_url)
    go_annotation_changes.write_json(output_rep + "go-ontology-changes.json", json_onto_changes)

    tsv_onto_changes = go_ontology_changes.create_text_report(json_onto_changes) 
    go_annotation_changes.write_text(output_rep + "go-ontology-changes.tsv", tsv_onto_changes)

    # with open('newstats/2019-06/go-ontology-changes.json', 'r') as myfile:
    #     data=myfile.read()
    # json_onto_changes = json.loads(data)
    print("DONE.")


    # 3 - Executing go_annotation_changes script
    print("\n\n3a - EXECUTING GO_ANNOTATION_CHANGES SCRIPT (INCLUDING PROTEIN BINDING)...\n")
    previous_stats = requests.get(previous_stats_url).json()    
    json_annot_changes = go_annotation_changes.compute_changes(json_stats, previous_stats)
    print("DONE.")
    
    print("\n\n3b - EXECUTING GO_ANNOTATION_CHANGES SCRIPT (EXCLUDING PROTEIN BINDING)...\n")
    previous_stats_no_pb = requests.get(previous_stats_no_pb_url).json()    # WE STILL NEED TO CORRECT THAT: 1 FILE OR SEVERAL FILE ? IF SEVERAL, ONE MORE PARAMETER
    json_annot_no_pb_changes = go_annotation_changes.compute_changes(json_stats_no_pb, previous_stats_no_pb)
    print("DONE.")

    # 4 - Refining go-stats with ontology stats
    print("\n\n4 - EXECUTING GO_REFINE_STATS SCRIPT...\n")
    merged_annotations_diff = merge_dict(json_stats, json_annot_changes)
    json_annot_changes = merged_annotations_diff


    ontology = json_onto_changes["summary"]["current"].copy()
    del ontology["release_date"]

    json_stats = {
        "release_date" : json_stats["release_date"],
        "ontology" : ontology,
        "annotations" : json_stats["annotations"]
    }
    go_stats.write_json(output_rep + "go-stats.json", json_stats)


    json_stats_no_pb = {
        "release_date" : json_stats_no_pb["release_date"],
        "ontology" : ontology,
        "annotations" : json_stats_no_pb["annotations"]
    }
    go_stats.write_json(output_rep + "go-stats-no-pb.json", json_stats_no_pb)


    annotations_by_reference_genome = { }
    for taxon in go_stats.reference_genomes_ids:
        key = go_stats.taxon_label(taxon)
        annotations_by_reference_genome[key] = json_stats["annotations"]["by_taxon"][key]

    bioentities_by_reference_genome = { }
    for taxon in go_stats.reference_genomes_ids:
        key = go_stats.taxon_label(taxon)
        bioentities_by_reference_genome[key] = json_stats["annotations"]["bioentities"]["by_taxon"]["cluster"][key]

    references_by_reference_genome = { }
    for taxon in go_stats.reference_genomes_ids:
        key = go_stats.taxon_label(taxon)
        references_by_reference_genome[key] = json_stats["annotations"]["references"]["all"]["by_taxon"][key]

    pmids_by_reference_genome = { }
    for taxon in go_stats.reference_genomes_ids:
        key = go_stats.taxon_label(taxon)
        pmids_by_reference_genome[key] = json_stats["annotations"]["references"]["pmids"]["by_taxon"][key]
        
    json_stats_summary = {
        "release_date" : json_stats["release_date"],
        "ontology" : ontology,
        "annotations" : {
            "total" : json_stats["annotations"]["total"],
            "total_no_pb" : json_stats_no_pb["annotations"]["total"],
            "by_evidence_cluster" : json_stats["annotations"]["by_evidence"]["cluster"],
            "by_evidence_cluster_no_pb" : json_stats_no_pb["annotations"]["by_evidence"]["cluster"],
            "by_reference_genome" : json_stats["annotations"]["by_reference_genome"]
        },
        "taxa" : {
            "total" : json_stats["annotations"]["taxa"]["total"],
            "filtered" : json_stats["annotations"]["taxa"]["filtered"],
        },
        "bioentities" : {
            "total" : json_stats["annotations"]["bioentities"]["total"],
            "total_no_pb" : json_stats_no_pb["annotations"]["bioentities"]["total"],
            "by_type_cluster" : json_stats["annotations"]["bioentities"]["by_type"]["cluster"],
            "by_type_cluster_no_pb" : json_stats_no_pb["annotations"]["bioentities"]["by_type"]["cluster"],
            "by_reference_genome" : bioentities_by_reference_genome
        },
        "references" : {
            "all" : {
                "total" : json_stats["annotations"]["references"]["all"]["total"],
                "total_no_pb" : json_stats_no_pb["annotations"]["references"]["all"]["total"],
                "by_reference_genome" : references_by_reference_genome
            }, 
            "pmids" : {
                "total" : json_stats["annotations"]["references"]["pmids"]["total"],
                "total_no_pb" : json_stats_no_pb["annotations"]["references"]["pmids"]["total"],
                "by_reference_genome" : pmids_by_reference_genome
            }
        },
    }

    # removing by_reference_genome.by_evidence
    for gen in json_stats_summary["annotations"]["by_reference_genome"]:
        del json_stats_summary["annotations"]["by_reference_genome"][gen]["by_evidence"]
    go_stats.write_json(output_rep + "go-stats-summary.json", json_stats_summary)


    # This is to modify the structure of the annotation changes based on recent requests
    json_annot_changes = alter_annotation_changes(json_stats, previous_stats, json_annot_changes)
    go_annotation_changes.write_json(output_rep + "go-annotation-changes.json", json_annot_changes)
    tsv_annot_changes = go_annotation_changes.create_text_report(json_annot_changes)
    go_annotation_changes.write_text(output_rep + "go-annotation-changes.tsv", tsv_annot_changes)

    json_annot_no_pb_changes = alter_annotation_changes(json_stats_no_pb, previous_stats_no_pb, json_annot_no_pb_changes)
    go_annotation_changes.write_json(output_rep + "go-annotation-changes_no_pb.json", json_annot_no_pb_changes)
    tsv_annot_changes_no_pb = go_annotation_changes.create_text_report(json_annot_no_pb_changes)
    go_annotation_changes.write_text(output_rep + "go-annotation-changes_no_pb.tsv", tsv_annot_changes_no_pb)


    print("SUCCESS.")




if __name__ == "__main__":
   main(sys.argv[1:])
   


# GO Update Statistics

import requests
import json
import sys, getopt, os

from xml.etree import ElementTree

# S3 specific imports
# import boto3
# from gzip import GzipFile
# from io import BytesIO

# s3_resource = boto3.resource('s3')
# go_s3_bucket = s3_resource.Bucket(name="geneontology-public")
# go_all_terms_key = "go-terms-aspect.json"
# go_stats_key= "go-stats"
# go_meta_key = "go-meta.json"
# go_most_annotated_gps_key= "go-annotated-gps.json"



# INPUT PARAMETERS
golr_base_url = 'http://golr-aux.geneontology.io/solr/'

ALL = "All"
BP = "GO:0008150"
MF = "GO:0003674"
CC = "GO:0005575"


# SET OF REFERENCE GENOMES (used to compute number of annotations by evidence code for each species)
reference_genomes_ids = [
    "NCBITaxon:9606",
    "NCBITaxon:10116",
    "NCBITaxon:10090",
    "NCBITaxon:3702",
    "NCBITaxon:7955",
    "NCBITaxon:6239",
    "NCBITaxon:559292",
    "NCBITaxon:7227",
    "NCBITaxon:44689",
    "NCBITaxon:4896",
    "NCBITaxon:83333"
]


# GOLR prepared queries
golr_select_ontology =  'select?wt=json&fq=document_category:"ontology_class"&fq=id:GO\:*&fq=idspace:"GO"&fl=source,annotation_class,is_obsolete&rows=2000000&q=*:*&facet=true&facet.field=source&facet.limit=1000000&facet.mincount=1'
golr_select_annotations = 'select?fq=document_category:%22annotation%22&q=*:*&wt=json&facet=true&facet.field=taxon&facet.field=aspect&facet.field=evidence_type&facet.field=assigned_by&facet.field=reference&facet.field=type&facet.limit=1000000&facet.mincount=1&rows=0'
# golr_select_annotations_no_pbinding = golr_select_annotations + "&fq=!isa_partof_closure:\"GO:0005515\"" # if we want to remove all derivates
golr_select_annotations_no_pbinding = golr_select_annotations + "&fq=!annotation_class:\"GO:0005515\"" # to remove only DIRECT annotations to protein binding
golr_select_bioentities = 'select?fq=document_category:%22bioentity%22&q=*:*&wt=json&facet=true&facet.field=type&facet.field=taxon&facet.limit=1000000&facet.mincount=1&rows=0'
golr_select_bioentities_pb = 'select?fq=document_category:"bioentity"&q=*:*&wt=json&rows=100000&fq=annotation_class_list:"GO:0005515"&fl=annotation_class_list,type,taxon'


def golr_fetch(select_query):
    # print("trying: " + golr_base_url + select_query)
    r = requests.get(golr_base_url + select_query)
    response = r.json()
    return response

def build_list(items_list, min_size = None):
    ls = []
    for i in range(0, len(items_list), 2):
        if min_size is None or items_list[i + 1] > min_size:
            ls.append(items_list[i])
    return ls

# utility function to transform a list [A, 1, B, 2] into a map {A: 1, B: 2}
def build_map(items_list, min_size = None):
    map = {}
    for i in range(0, len(items_list), 2):
        if min_size is None or items_list[i + 1] > min_size:
            map[items_list[i]] = items_list[i + 1]
    return map

# utility function to build a reverse map: { "a": 1, "b": 1, "c": 2 } -> {1: ["a", "b"], 2: ["c"]}
def build_reverse_map(map):
    reverse_map = { }
    for key, val in map.items():
        ls = []
        if val in reverse_map:
            ls = reverse_map[val]
        else:
            reverse_map[val] = ls
        ls.append(key)
    return reverse_map

# utility function to cluster elements of an input map based on another map of synonyms
def cluster_map(input_map, synonyms):
    cluster = { }
    for key, val in input_map.items():
        temp = synonyms[key]
        if temp in cluster:
            val_cluster = cluster[temp]
            cluster[temp] = val_cluster + val
        else:
            cluster[temp] = val
    return cluster

# similar as above but the value of each key is also a map
def cluster_complex_map(input_map, synonyms):
    cluster = { }
    for key, val in input_map.items():
        temp = synonyms[key]
        # print("working on : " , key , val)
        if temp in cluster:
            temp_cluster = cluster[temp]
            # print("cluster already found : ", temp , temp_cluster)
            for key_cluster, val_cluster in temp_cluster.items():
                temp_cluster[key_cluster] = val_cluster + val[key_cluster]
        else:
            cluster[temp] = val
    return cluster


# reorder map (python 3.6 keeps order in which items are inserted in map: https://stackoverflow.com/questions/613183/how-do-i-sort-a-dictionary-by-value)
def ordered_map(map):
    ordered_map = { }
    for w in sorted(map, key=map.get, reverse=True):
        ordered_map[w] = map[w]
    return ordered_map
    
def extract_map(map, key_str):
    extracted = { }
    for key, val in map.items():
        if key_str in key:
            extracted[key] = val
    return extracted

# useful grouping of evidences as discussed with Pascale
evidence_groups = {
    "EXP": ["EXP", "IDA", "IEP", "IGC", "IGI", "IMP", "IPI"],
    "HTP": ["HDA", "HEP", "HGI", "HMP", "HTP"],
    "PHYLO": ["IBA", "IRD", "IKR", "IMR"],
    "IEA": ["IEA"],
    "ND": ["ND"],
    "OTHER": ["IC", "ISA", "ISM", "ISO", "ISS", "NAS", "RCA", "TAS"]
}

# convenience reverse map of evidence groups
reverse_evidence_groups = { }
for key, value in evidence_groups.items():                                                                                                                                                    
    for key2 in value:                                                                                                                                                                        
        reverse_evidence_groups[key2] = key    

# auto computed set of ~ 200 species with > 1000 annotations
usable_taxons = [ ]

# fetch at startup from eutils.ncbi.nlm and parse into a map
taxon_base_url = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=taxonomy'
taxon_map_fallback_url = 'https://geneontology.s3.amazonaws.com/taxon_map.json'
taxon_map = { }

# auto computed set of groups doing annotations (assigned_by)
groups = [ ]

# auto computed set of bioentity types
bioentity_types = [ ]

# auto computed map to collapse RNA under a same cluster
bioentity_type_cluster = { }

# convenience reverse map of bioentity_type_cluster
reverse_bioentity_type_cluster = { }

# will contain the last release date from following url
# release_date = "N/A"
# go_pipeline_release_url = "http://current.geneontology.org/metadata/release-date.json"

# def get_release_date():
    # r = requests.get(go_pipeline_release_url)
    # return r.json()['date']






# def lambda_handler(event, context):
    # stats = compute_stats(golr_base_url)
    # return stats

def compute_stats(golr_url, release_date, exclude_pb_only = False):
    """
    compute stats on GO annotations - can specify if we include or exclude annotations to protein binding only
    """
    global golr_base_url
    golr_base_url = golr_url

    print("Will use golr url: " , golr_base_url)

    # global release_date
    # release_date = get_release_date()

    print("1 / 4 - Fetching GO terms...")
    all_terms = golr_fetch(golr_select_ontology)
    print("Done.")
    
    print("2 / 4 - Fetching GO annotations...")
    if exclude_pb_only:
        all_annotations = golr_fetch(golr_select_annotations_no_pbinding)
    else:
        all_annotations = golr_fetch(golr_select_annotations)
    print("Done.")
    
    print("3 / 4 - Fetching GO bioentities...")
    all_entities = golr_fetch(golr_select_bioentities)

    # we have to manually update the facts of the first query if we want to remove the bioentities annotated only to protein binding
    if exclude_pb_only:
        all_entities_no_pb = golr_fetch(golr_select_bioentities_pb)
        # print(all_entities_no_pb)
        entities_type_no_pb = { }
        entities_taxon_no_pb = { }

        count = 0

        for doc in all_entities_no_pb['response']['docs']:
            if len(doc['annotation_class_list']) > 1:
                continue
            count += 1
            if doc['type'] in entities_type_no_pb:
                entities_type_no_pb[doc['type']] += 1
            else:
                entities_type_no_pb[doc['type']] = 1

            if doc['taxon'] in entities_type_no_pb:
                entities_taxon_no_pb[doc['taxon']] += 1
            else:
                entities_taxon_no_pb[doc['taxon']] = 1

        # finally update the type facet field
        types = all_entities['facet_counts']['facet_fields']['type']
        for i in range(0, len(types), 2):
            ctype = types[i]
            retr_value = entities_type_no_pb[ctype] if ctype in entities_type_no_pb else 0
            types[i + 1] = types[i + 1] - retr_value
        all_entities['facet_counts']['facet_fields']['type'] = types

        all_entities['response']['numFound'] = all_entities['response']['numFound'] - count
       
        # and update the taxon facet field
        taxons = all_entities['facet_counts']['facet_fields']['taxon']
        for i in range(0, len(taxons), 2):
            ctaxon = taxons[i]
            retr_value = entities_taxon_no_pb[ctaxon] if ctaxon in entities_taxon_no_pb else 0
            taxons[i + 1] = taxons[i + 1] - retr_value
        all_entities['facet_counts']['facet_fields']['taxon'] = taxons
    
    print("Done.")

    print("4 / 4 - Creating Stats...")    
    prepare_globals(all_annotations)
    print("\t4a - globals prepared")
    stats = create_stats(all_terms, all_annotations, all_entities, release_date)
    print("Done.")
    
    return stats

def load_taxon_map():
    global taxon_map
    print("Using ", taxon_map_fallback_url , " (created from ftp://ftp.ncbi.nlm.nih.gov/pub/taxonomy/taxdmp.zip) as a fallback to get taxon { id, label }")
    data = requests.get(taxon_map_fallback_url)

    if data.status_code != 200:
        return False

    taxon_map = json.loads(data.content)
    check = taxon_map['9606'] == 'Homo sapiens'
    return check

def prepare_globals(all_annotations):
    global usable_taxons
    global taxon_map
    global groups
    global bioentity_types
    global bioentity_type_cluster
    global reverse_bioentity_type_cluster    

    temp = all_annotations['facet_counts']['facet_fields']['assigned_by']
    groups = build_list(temp)

    temp = all_annotations['facet_counts']['facet_fields']['taxon']
    usable_taxons = build_list(temp, 1000)
    all_taxons = build_list(temp, None)

    # this step will create the global taxon_map to get any name from an id
    temp_taxons = []
    for taxon in all_taxons:
        temp_taxons.append(taxon[taxon.index(":")+1:])

    params = { "id" : ",".join(temp_taxons) }
    data = requests.post(taxon_base_url, data = params)

    if data.status_code == 200:
        tree = ElementTree.fromstring(data.content)
        elts = tree.findall("Taxon")
        for i in range(0,len(elts)):
            key = elts[i].findtext("TaxId")
            val = elts[i].findtext("ScientificName")
            taxon_map[key] = val
        print("Note: taxon map of ", len(taxon_map), " taxa loaded from " , taxon_base_url + " - in case of issue could use https://www.ebi.ac.uk/ena/data/taxonomy/v1/taxon/tax-id/xxx")
    else:
        print("WARNING: could not get taxon labels from ", taxon_base_url , " (status code: " , str(data.status_code) + ")")
        load_taxon_map()

    # verbose check on taxon label mapping
    check = taxon_map['9606'] == 'Homo sapiens'
    if check:
        print("Successfully pass taxon label mapping test (taxon_map['9606'] == 'Homo sapiens'): ", taxon_map['9606'] == 'Homo sapiens')
    else:
        print("Taxon map could not be created, will show taxon labels as UNK")
        

    bioentity_type_cluster = { }
    temp = all_annotations['facet_counts']['facet_fields']['type']
    for i in range(0, len(temp), 2):
        bioentity_types.append(temp[i])
        bioentity_type_cluster[temp[i]] = bioentity_type(temp[i])

    reverse_bioentity_type_cluster = build_reverse_map(bioentity_type_cluster)

def bioentity_type(str_type):
    """
    In a nutshell, collapse all RNA related types into RNA
    """
    if "RNA" in str_type or "ribozyme" in str_type or "transcript" in str_type:
        return "RNA"
    return str_type
    
def golr_fetch_bioentities_taxon(taxon):
    url = "select?fq=document_category:%22bioentity%22&q=*:*&wt=json&rows=0&facet=true&facet.field=type&facet.field=taxon&facet.limit=1000000&facet.mincount=1&fq=taxon:\"" + taxon + "\""
    response = golr_fetch(url)

    # multiple queries: a bit complicated but necessary due to solr 3.6 unable to do composite faceting and for speed considerations
    # * can indicate the is_a closure to find the stats on that specific aspect
    # * if evidence code was present, we could use a similar strategy
    url_bp = "select?fq=document_category:%22bioentity%22&q=*:*&wt=json&facet=true&facet.field=type&facet.field=taxon&facet.limit=1000000&facet.mincount=1&rows=0&fq=taxon:\"" + taxon + "\"&fq=isa_partof_closure:\"" + BP + "\""
    response_bp = golr_fetch(url_bp)

    url_mf = "select?fq=document_category:%22bioentity%22&q=*:*&wt=json&facet=true&facet.field=type&facet.field=taxon&facet.limit=1000000&facet.mincount=1&rows=0&fq=taxon:\"" + taxon + "\"&fq=isa_partof_closure:\"" + MF + "\""
    response_mf = golr_fetch(url_mf)

    url_cc = "select?fq=document_category:%22bioentity%22&q=*:*&wt=json&facet=true&facet.field=type&facet.field=taxon&facet.limit=1000000&facet.mincount=1&rows=0&fq=taxon:\"" + taxon + "\"&fq=isa_partof_closure:\"" + CC + "\""
    response_cc = golr_fetch(url_cc)

    return { ALL : response, BP : response_bp, MF : response_mf, CC : response_cc }
    
def golr_fetch_references_taxon(taxon):
    url = "select?fq=document_category:%22annotation%22&q=*:*&wt=json&rows=0&facet.limit=10000000&facet.mincount=1&facet=true&facet.field=reference&fq=taxon:\"" + taxon + "\""
    response = golr_fetch(url)
    return response

def golr_fetch_references_group(group):
    url = "select?fq=document_category:%22annotation%22&q=*:*&wt=json&rows=0&facet.limit=10000000&facet.mincount=1&facet=true&facet.field=reference&fq=assigned_by:\"" + group + "\""
    response = golr_fetch(url)
    return response

# def golr_fetch_evidence_by_species(taxon_id):
#     url = 'select?fq=document_category:%22annotation%22&q=*:*&wt=json&fq=taxon:%22' + taxon_id + '%22&facet=true&facet.field=evidence_type&facet.limit=10000&rows=0'
#     response = golr_fetch(url)
#     return response    


def golr_fetch_evidence_by_species(taxon):
    url = 'select?fq=document_category:%22annotation%22&q=*:*&wt=json&fq=taxon:%22' + taxon + '%22&facet=true&facet.field=evidence_type&facet.limit=10000&rows=0'
    response = golr_fetch(url)

    url_bp = 'select?fq=document_category:%22annotation%22&q=*:*&wt=json&fq=taxon:%22' + taxon + '%22&facet=true&facet.field=evidence_type&facet.limit=10000&rows=0&fq=isa_partof_closure:\"' + BP + '\"'
    response_bp = golr_fetch(url_bp)

    url_mf = 'select?fq=document_category:%22annotation%22&q=*:*&wt=json&fq=taxon:%22' + taxon + '%22&facet=true&facet.field=evidence_type&facet.limit=10000&rows=0&fq=isa_partof_closure:\"' + MF + '\"'
    response_mf = golr_fetch(url_mf)

    url_cc = 'select?fq=document_category:%22annotation%22&q=*:*&wt=json&fq=taxon:%22' + taxon + '%22&facet=true&facet.field=evidence_type&facet.limit=10000&rows=0&fq=isa_partof_closure:\"' + CC + '\"'
    response_cc = golr_fetch(url_cc)

    return { ALL : response, BP : response_bp, MF : response_mf, CC : response_cc }
    


def taxon_label(taxon):
    if "NCBITaxon" in taxon:
        taxon_id = taxon[taxon.index(":")+1:]
        taxon_name = taxon_map[taxon_id] if taxon_id in taxon_map else "UNK"
        return taxon + "|" + taxon_name
    return taxon

def add_taxon_label(map):
    new_map = { }
    for key, val in map.items():
        if "NCBITaxon" in key:
            taxon_id = key[key.index(":")+1:]
            taxon_name = taxon_map[taxon_id] if taxon_id in taxon_map else "UNK" 
            # print("mapping ", taxon_id)
            if type(val) == dict:
                new_map[key + "|" + taxon_name] = add_taxon_label(val)
            else:
                new_map[key + "|" + taxon_name] = val
        else:
            if type(val) == dict:
                new_map[key] = add_taxon_label(val)
            else:
                new_map[key] = val
    return new_map

def create_stats(all_terms, all_annotations, all_entities, release_date):
    stats = { }

    terms = 0
    obsoleted = 0
    terms_by_aspect = { "P" : 0, "F" : 0, "C" : 0 }

    for doc in all_terms['response']['docs']:
        if doc['is_obsolete']:
            obsoleted += 1
        else:
            terms += 1

            # some obsoleted annotations don't have a source
            if 'source' not in doc:
                continue

            if "biological_process" in doc['source']:
                terms_by_aspect["P"] += 1
            if "molecular_function" in doc['source']:
                terms_by_aspect["F"] += 1
            if "cellular_component" in doc['source']:
                terms_by_aspect["C"] += 1
            
    terms = { 
        "total" : all_terms['response']['numFound'],
        "valid" : terms,
        "obsoleted" : obsoleted,
        "by_aspect" : terms_by_aspect
    }
    print("\t4b - terms computed")

    all_bioentities_by_taxon = { }
    cluster_bioentities_by_taxon = { }
    for taxon in usable_taxons:
        responses = golr_fetch_bioentities_taxon(taxon)
        all_map = build_map(responses[ALL]['facet_counts']['facet_fields']['type'])
        bp_map = build_map(responses[BP]['facet_counts']['facet_fields']['type'])
        mf_map = build_map(responses[MF]['facet_counts']['facet_fields']['type'])
        cc_map = build_map(responses[CC]['facet_counts']['facet_fields']['type'])

        merged_map = {}
        for key, value in all_map.items():
            merged_map[key] = { "A" : value , "P" : bp_map[key] if key in bp_map else 0 , "F" : mf_map[key] if key in mf_map else 0 , "C" : cc_map[key] if key in cc_map else 0 }

        all_bioentities_by_taxon[taxon] = merged_map
        cluster_bioentities_by_taxon[taxon] =  cluster_complex_map(all_bioentities_by_taxon[taxon], bioentity_type_cluster)
        
        # all_bioentities_by_taxon[taxon] = build_map(res['facet_counts']['facet_fields']['type'])
        # cluster_bioentities_by_taxon[taxon] =  cluster_map(all_bioentities_by_taxon[taxon], bioentity_type_cluster)
    print("\t4c - bioentities computed")

    references_by_taxon = { }
    pmids_by_taxon = { }
    for taxon in usable_taxons:
        res = golr_fetch_references_taxon(taxon)
        references_by_taxon[taxon] = int(len(res['facet_counts']['facet_fields']['reference']) / 2)
        pmid_map = build_map(res['facet_counts']['facet_fields']['reference'])
        pmid_map = len(extract_map(pmid_map, "PMID:"))
        pmids_by_taxon[taxon] = pmid_map
    references_by_taxon = ordered_map(references_by_taxon)
    pmids_by_taxon = ordered_map(pmids_by_taxon)
    print("\t4d - taxa computed")

    references_by_group = { }
    pmids_by_group = { }
    for group in groups:
        res = golr_fetch_references_group(group)
        references_by_group[group] = int(len(res['facet_counts']['facet_fields']['reference']) / 2)
        pmid_map = build_map(res['facet_counts']['facet_fields']['reference'])
        pmid_map = len(extract_map(pmid_map, "PMID:"))
        pmids_by_group[group] = pmid_map
    references_by_group = ordered_map(references_by_group)
    pmids_by_group = ordered_map(pmids_by_group)
    print("\t4e - references computed")

    # print("CHECK (by evidence):\n" , all_annotations['facet_counts']['facet_fields']['evidence_type'])
    # print("CHECK (buildmap(by_evidence):\n", build_map(all_annotations['facet_counts']['facet_fields']['evidence_type']))
    # print("CHECK (reverse_evidence_group:\n", reverse_evidence_groups)
    # print("CHECK (cluster(buildmap, reverse_evidence_group):\n", cluster_map(build_map(all_annotations['facet_counts']['facet_fields']['evidence_type']), reverse_evidence_groups))




    ref_genome_evidences = { }
    for taxon in reference_genomes_ids:
        responses = golr_fetch_evidence_by_species(taxon)
        all_map = build_map(responses[ALL]['facet_counts']['facet_fields']['evidence_type'])
        bp_map = build_map(responses[BP]['facet_counts']['facet_fields']['evidence_type'])
        mf_map = build_map(responses[MF]['facet_counts']['facet_fields']['evidence_type'])
        cc_map = build_map(responses[CC]['facet_counts']['facet_fields']['evidence_type'])

        merged_map = {}
        for key, value in all_map.items():
            merged_map[key] = { "A" : value , "P" : bp_map[key] if key in bp_map else 0 , "F" : mf_map[key] if key in mf_map else 0 , "C" : cc_map[key] if key in cc_map else 0 }

        ref_genome_evidences[taxon] = {
            "by_evidence" : merged_map
        }
        ref_genome_evidences[taxon]["by_evidence_cluster"] = cluster_complex_map(ref_genome_evidences[taxon]["by_evidence"], reverse_evidence_groups)
        

    annotations = { 
        "total" : all_annotations['response']['numFound'],

        "bioentities" : {
            "total" : all_entities['response']['numFound'],

            "by_type" : {
                "all" : build_map(all_entities['facet_counts']['facet_fields']['type']),
                "cluster" : cluster_map(build_map(all_entities['facet_counts']['facet_fields']['type']), bioentity_type_cluster)
            },

            "by_taxon" : {
                "all" : all_bioentities_by_taxon,
                "cluster" : cluster_bioentities_by_taxon
            }

            # This can not work and would require an evidence fields in the GOLR bioentity docs
            # "by_taxon" : {
            #     "all" : all_bioentities_by_taxon,
            #     "experimental" : experimental_bioentities_by_taxon
            # }
        },

        "taxa" : {
            "total" : int(len(all_annotations['facet_counts']['facet_fields']['taxon']) / 2),
            "filtered" : len(usable_taxons),
        },

        "references" : {
            "all" : {
                "total" : int(len(all_annotations['facet_counts']['facet_fields']['reference']) / 2),
                "by_taxon" : references_by_taxon,
                "by_group" : references_by_group
            },
            "pmids" : {
                "total" : len(extract_map(build_map(all_annotations['facet_counts']['facet_fields']['reference']), "PMID:")),
                "by_taxon" : pmids_by_taxon,
                "by_group" : pmids_by_group
            }
        },


        "by_aspect" : build_map(all_annotations['facet_counts']['facet_fields']['aspect']),

        "by_bioentity_type" : {
            "all" : build_map(all_annotations['facet_counts']['facet_fields']['type']),
            "cluster" : cluster_map(build_map(all_annotations['facet_counts']['facet_fields']['type']), bioentity_type_cluster)
        },

        "by_taxon": build_map(all_annotations['facet_counts']['facet_fields']['taxon']),

        "by_evidence": {
            "all" : build_map(all_annotations['facet_counts']['facet_fields']['evidence_type']),
            "cluster" : cluster_map(build_map(all_annotations['facet_counts']['facet_fields']['evidence_type']), reverse_evidence_groups)
        },

        "by_reference_genome" : ref_genome_evidences,

        "by_group": build_map(all_annotations['facet_counts']['facet_fields']['assigned_by'])
        
    }

    annotations = add_taxon_label(annotations)

    stats["release_date"] = release_date
    stats["terms"] = terms
    stats["annotations"] = annotations

    return stats


def create_text_report(stats_json):
    text_report = ""

    text_report = "GENE ONTOLOGY STATISTICS"
    text_report += "\nrelease_date\t" + stats_json["release_date"]    

    text_report += "\n\nTERMS\n"
    text_report += "total\t" + str(stats_json["terms"]["total"]) + "\nobsoleted\t" + str(stats_json["terms"]["obsoleted"]) + "\nvalid total\t" + str(stats_json["terms"]["valid"])
    text_report += "\nvalid P\t" + str(stats_json["terms"]["by_aspect"]["P"]) + "\nvalid F\t" + str(stats_json["terms"]["by_aspect"]["F"]) + "\nvalid C\t" + str(stats_json["terms"]["by_aspect"]["C"])


    text_report += "\n\nBIOENTITIES\n"
    text_report += "total\t" + str(stats_json["annotations"]["bioentities"]["total"])

    text_report += "\n\nBIOENTITIES BY TYPE (CLUSTER)"
    for key, val in stats_json["annotations"]["bioentities"]["by_type"]["cluster"].items():
        text_report += "\n" + key + "\t" + str(val)

    text_report += "\n\nBIOENTITIES BY TYPE (ALL)"
    for key, val in stats_json["annotations"]["bioentities"]["by_type"]["all"].items():
        text_report += "\n" + key + "\t" + str(val)

    text_report += "\n\nBIOENTITIES BY FILTERED TAXON AND BY TYPE (CLUSTER)"
    text_report += "\ntaxon"
    for type, nb in stats_json["annotations"]["bioentities"]["by_type"]["cluster"].items():
        text_report += "\t" + type
    for key, val in stats_json["annotations"]["bioentities"]["by_taxon"]["cluster"].items():
        text_report += "\n" + key
        for type, nb in stats_json["annotations"]["bioentities"]["by_type"]["cluster"].items():
            text_report += "\t" + str(val[type]["A"]) if type in val else "\t0"

    text_report += "\n\nBIOENTITIES BY FILTERED TAXON AND BY TYPE (ALL)"
    text_report += "\ntaxon"
    for type, nb in stats_json["annotations"]["bioentities"]["by_type"]["all"].items():
        text_report += "\t" + type
    for key, val in stats_json["annotations"]["bioentities"]["by_taxon"]["all"].items():
        text_report += "\n" + key
        for type, nb in stats_json["annotations"]["bioentities"]["by_type"]["all"].items():
            text_report += "\t" + str(val[type]["A"]) if type in val else "\t0"


    text_report += "\n\nTAXA\n"
    text_report += "total\t" + str(stats_json["annotations"]["taxa"]["total"]) + "\nfiltered\t" + str(stats_json["annotations"]["taxa"]["filtered"])


    text_report += "\n\nANNOTATIONS\n"
    text_report += "total\t" + str(stats_json["annotations"]["total"])
    for key, val in stats_json["annotations"]["by_aspect"].items():
        text_report += "\n" + key + "\t" + str(val)
        
    text_report += "\n\nANNOTATIONS BY BIOENTITY TYPE (CLUSTER)"
    for key, val in stats_json["annotations"]["by_bioentity_type"]["cluster"].items():
        text_report += "\n" + key + "\t" + str(val)
    
    text_report += "\n\nANNOTATIONS BY BIOENTITY TYPE (ALL)"
    for key, val in stats_json["annotations"]["by_bioentity_type"]["all"].items():
        text_report += "\n" + key + "\t" + str(val)
    
    text_report += "\n\nANNOTATIONS BY EVIDENCE (CLUSTER)"
    for key, val in stats_json["annotations"]["by_evidence"]["cluster"].items():
        text_report += "\n" + key + "\t" + str(val)
    
    text_report += "\n\nANNOTATIONS BY EVIDENCE (ALL)"
    for key, val in stats_json["annotations"]["by_evidence"]["all"].items():
        text_report += "\n" + key + "\t" + str(val)
    
    text_report += "\n\nANNOTATIONS BY GROUP"
    for key, val in stats_json["annotations"]["by_group"].items():
        text_report += "\n" + key + "\t" + str(val)
    
    text_report += "\n\nANNOTATIONS BY TAXON"
    for key, val in stats_json["annotations"]["by_taxon"].items():
        text_report += "\n" + key + "\t" + str(val)


    text_report += "\n\nREFERENCES AND PMIDS\n"
    text_report += "total\t" + str(stats_json["annotations"]["references"]["all"]["total"]) + "\t" + str(stats_json["annotations"]["references"]["pmids"]["total"])

    text_report += "\n\nREFERENCES AND PMIDS BY GROUP"
    text_report += "\ngroup\treferences\tpmids"
    for key, val in stats_json["annotations"]["references"]["all"]["by_group"].items():
        text_report += "\n" + key + "\t" + str(val) + "\t" + str(stats_json["annotations"]["references"]["pmids"]["by_group"][key])

    text_report += "\n\nREFERENCES AND PMIDS BY TAXON"
    text_report += "\ntaxon\treferences\tpmids"
    for key, val in stats_json["annotations"]["references"]["all"]["by_taxon"].items():
        pmid_val = stats_json["annotations"]["references"]["pmids"]["by_taxon"][key] if key in stats_json["annotations"]["references"]["pmids"]["by_taxon"] else 0
        text_report += "\n" + key + "\t" + str(val) + "\t" + str(pmid_val)

    return text_report

def create_meta(json_stats):
    meta = {
        "release_date": json_stats["release_date"],
        "terms": {
            "total" : json_stats["terms"]["valid"],
            "by_aspect" : json_stats["terms"]["by_aspect"]
        },
        "annotations" : {
            "total" : json_stats["annotations"]["total"],
            "experimental" : json_stats["annotations"]["by_evidence"]["cluster"]["EXP"]
        },
        "taxa": {
            "total" : json_stats["annotations"]["taxa"]["total"],
            "filtered" : json_stats["annotations"]["taxa"]["filtered"]
        },
        "bioentities" : {
            "total" : json_stats["annotations"]["bioentities"]["total"],
            "by_type_cluster" :json_stats["annotations"]["bioentities"]["by_type"]["cluster"]
        },
        "references": {
            "all" : json_stats["annotations"]["references"]["all"]["total"],
            "pmids": json_stats["annotations"]["references"]["pmids"]["total"]
        }
    }
    return meta
    
# def save_terms_mapping(all_terms):
#     for doc in all_terms['response']['docs']:
#         doc.pop('is_obsolete', None)

#     # creating the results divided by aspect        
#     mapping = { }
#     for term in all_terms['response']['docs']:
#         if 'source' not in term: # often, obsoleted terms have no source
#             print("Term " , term , " has no source !")
#             continue
#         if term['source'] not in mapping:
#             mapping[term['source']] = []
#         mapping[term['source']].append(term['annotation_class'])
    
#     store_json(go_all_terms_key, mapping)
#     store_json("archive/" + release_date + "_" + go_all_terms_key, mapping)
 

# def store_text(key, content):
#     # Storing a compressed version of the text file
#     gz_body = BytesIO()
#     gz = GzipFile(None, 'wb', 9, gz_body)
#     gz.write(content.encode('utf-8'))
#     gz.close()    

#     go_s3_bucket.put_object(
#         Key=key,  
#         ContentType='text/plain', 
#         ContentEncoding='gzip', 
#         Body=gz_body.getvalue()
#     )    
    
# def store_json(key, content):
#     # Storing a compressed version of the json file
#     gz_body = BytesIO()
#     gz = GzipFile(None, 'wb', 9, gz_body)
#     gz.write(json.dumps(content).encode('utf-8'))
#     gz.close()    

#     go_s3_bucket.put_object(
#         Key=key,  
#         ContentType='application/json', 
#         ContentEncoding='gzip', 
#         Body=gz_body.getvalue()
#     )    

def write_json(key, content):
    with open(key, 'w') as outfile:
        json.dump(content, outfile, indent=2)
 
def write_text(key, content):
    with open(key, 'w') as outfile:
        outfile.write(content)



def print_help():
    print('\nUsage: python go_stats.py -g <golr_url> -d <release_date> -o <output_rep>\n')


def main(argv):
    golr_url = ''
    output_rep = ''
    release_date = ''

    if len(argv) < 6:
        print_help()
        sys.exit(2)

    try:
        opts, argv = getopt.getopt(argv,"g:b:o:d:",["golrurl=","orep=","date="])
    except getopt.GetoptError:
        print_help()
        sys.exit(2)

    for opt, arg in opts:
        if opt == '-h':
            print_help()
            sys.exit()
        elif opt in ("-g", "--golrurl"):
            golr_url = arg
        elif opt in ("-o", "--orep"):
            output_rep = arg
        elif opt in ("-d", "--date"):
            release_date = arg

    if not output_rep.endswith("/"):
        output_rep += "/"

    if not os.path.exists(output_rep):
        os.mkdir(output_rep)

    output_meta = output_rep + "go-meta.json"
    output_meta_no_pb = output_rep + "go-meta-no-pb.json"
    output_json =  output_rep + "go-stats.json"
    output_no_pb_json =  output_rep + "go-stats-no-pb.json"
    output_tsv =  output_rep + "go-stats.tsv"
    output_no_pb_tsv =  output_rep + "go-stats-no-pb.tsv"


    print("Will write stats to " + output_json + " and " + output_tsv)
    json_stats = compute_stats(golr_url, release_date, False)
    print("Saving Stats to <" + output_json + "> ...")    
    write_json(output_json, json_stats)
    print("Done.")

    print("Saving Stats to <" + output_tsv + "> ...")    
    tsv_stats = create_text_report(json_stats)
    write_text(output_tsv, tsv_stats)
    print("Done.")


    print("Will write stats (excluding protein binding) to " + output_no_pb_json + " and " + output_no_pb_tsv)
    json_stats_no_pb = compute_stats(golr_url, release_date, True)
    print("Saving Stats to <" + output_no_pb_json + "> ...")    
    write_json(output_no_pb_json, json_stats_no_pb)
    print("Done.")

    print("Saving Stats (excluding protein binding) to <" + output_no_pb_tsv + "> ...")    
    tsv_stats_no_pb = create_text_report(json_stats_no_pb)
    write_text(output_no_pb_tsv, tsv_stats_no_pb)
    print("Done.")


    json_meta = create_meta(json_stats)
    print("Saving META to <" + output_meta + "> ...")    
    write_json(output_meta, json_meta)
    print("Done.")


    json_meta_no_pb = create_meta(json_stats_no_pb)
    print("Saving META to <" + output_meta_no_pb + "> ...")    
    write_json(output_meta_no_pb, json_meta_no_pb)
    print("Done.")
    


if __name__ == "__main__":
   main(sys.argv[1:])
   


import sys, getopt, os

import requests
import json



def compute_changes(current_stats, previous_stats):
    stats_changes = {
        "releases_compared": {
            "current" : current_stats['release_date'],
            "previous" : previous_stats['release_date']            
        }
    }

    for key, val in current_stats.items():
        if "release_date" in key:
            continue
        if "terms" in key:
            continue
        stats_changes[key] = nested_changes(current_stats[key], previous_stats[key])

    return stats_changes


def nested_changes(current_json, previous_json):
    """
    Compute nested changes of numbers (ignore string).
    Assume both json have the exact same structure
    """

    changes = { }

    for key, val in current_json.items():
        tp = type(val)
        if tp == str:
            continue
        elif tp == int or tp == float:
            previous_value = previous_json[key] if previous_json is not None and key in previous_json else 0
            # if current_json[key] != previous_value:
            changes[key] = current_json[key] - previous_value
        elif tp == dict:
            previous_value = previous_json[key] if previous_json is not None and key in previous_json else { }
            changes[key] = nested_changes(current_json[key], previous_value)

    missing = missing_fields(current_json, previous_json)
    for key, val in missing.items():
        changes[key] = val

    return changes

def missing_fields(current_json, previous_json):
    """ 
    Adding fields that were in previous but are not in current stats
    """
    missing = { }
    for key, val in previous_json.items():
        if current_json is not None and key in current_json:
            continue

        tp = type(val)
        if tp == str:
            continue
        elif tp == int or tp == float:
            missing[key] = - val
        elif tp == dict:
            missing[key] = missing_fields(None, previous_json[key])
            
    return missing

def create_text_report(json_changes):
    text_report = ""

    text_report = "CHANGES IN GO ANNOTATIONS"
    text_report += "\ncurrent_release_date\t" + json_changes["releases_compared"]["current"]    
    text_report += "\nprevious_release_date\t" + json_changes["releases_compared"]["previous"]

    # text_report += "\n\nCHANGES IN TERMS\n"
    # text_report += "total\t" + str(json_changes["terms"]["total"]) + "\nobsoleted\t" + str(json_changes["terms"]["obsoleted"]) + "\nvalid total\t" + str(json_changes["terms"]["valid"])
    # text_report += "\nvalid P\t" + str(json_changes["terms"]["by_aspect"]["P"]) + "\nvalid F\t" + str(json_changes["terms"]["by_aspect"]["F"]) + "\nvalid C\t" + str(json_changes["terms"]["by_aspect"]["C"])

    text_report += "\n\nSUMMARY: CURRENT_RELEASE (" + json_changes["releases_compared"]["current"] + ")"
    text_report += "\nannotated bioentities:\t" + str(json_changes["summary"]["current"]["bioentities"])
    text_report += "\nannotated taxa:\t" + str(json_changes["summary"]["current"]["taxa"]) + "\t" + str(json_changes["summary"]["current"]["taxa_filtered"]) + " (with more than 1000 annotations)"
    text_report += "\nannotations:\t" + str(json_changes["summary"]["current"]["annotations"]["total"])
    for key, val in json_changes["summary"]["current"]["annotations"]["by_aspect"].items():
        text_report += "\nannotations by aspect " + key + ":\t" + str(val)
    for key, val in json_changes["summary"]["current"]["annotations"]["by_evidence_cluster"].items():
        text_report += "\nannotations by evidence cluster " + key + ":\t" + str(val)

    text_report += "\n\nSUMMARY: PREVIOUS (" + json_changes["releases_compared"]["previous"] + ")"
    text_report += "\nannotated bioentities:\t" + str(json_changes["summary"]["previous"]["bioentities"])
    text_report += "\nannotated taxa:\t" + str(json_changes["summary"]["previous"]["taxa"]) + "\t" + str(json_changes["summary"]["previous"]["taxa_filtered"]) + " (with more than 1000 annotations)"
    text_report += "\nannotations:\t" + str(json_changes["summary"]["previous"]["annotations"]["total"])
    for key, val in json_changes["summary"]["previous"]["annotations"]["by_aspect"].items():
        text_report += "\nannotations by aspect " + key + ":\t" + str(val)
    for key, val in json_changes["summary"]["previous"]["annotations"]["by_evidence_cluster"].items():
        text_report += "\nannotations by evidence cluster " + key + ":\t" + str(val)

    text_report += "\n\nSUMMARY: DIFF BETWEEN RELEASES"
    text_report += "\nannotated bioentities:\t" + str(json_changes["summary"]["current"]["bioentities"] - json_changes["summary"]["previous"]["bioentities"])
    text_report += "\nannotated taxa:\t" + str(json_changes["summary"]["current"]["taxa"] - json_changes["summary"]["previous"]["taxa"]) + "\t" + str(json_changes["summary"]["current"]["taxa_filtered"] - json_changes["summary"]["previous"]["taxa_filtered"]) + " (with more than 1000 annotations)"
    text_report += "\nannotations:\t" + str(json_changes["summary"]["current"]["annotations"]["total"] - json_changes["summary"]["previous"]["annotations"]["total"])
    for key, val in json_changes["summary"]["current"]["annotations"]["by_aspect"].items():
        text_report += "\nannotations by aspect " + key + ":\t" + str(val - json_changes["summary"]["previous"]["annotations"]["by_aspect"][key])
    # for key, val in json_changes["summary"]["current"]["annotations"]["by_evidence_cluster"].items():
    #     text_report += "\nannotations by evidence cluster " + key + ":\t" + str(val - json_changes["summary"]["current"]["annotations"]["by_evidence_cluster"][key])
    for key, val in json_changes["summary"]["changes"]["annotations"]["by_evidence_cluster"].items():
        text_report += "\nannotations by evidence cluster " + key + ":\t" + str(json_changes["summary"]["changes"]["annotations"]["by_evidence_cluster"][key])

    text_report += "\n\nDETAILED CHANGES"

    text_report += "\n\nCHANGES IN ANNOTATED BIOENTITIES\n"
    text_report += "total\t" + str(json_changes["detailed_changes"]["bioentities"]["total"])

    text_report += "\n\nCHANGES IN ANNOTATED BIOENTITIES BY TYPE (CLUSTER)"
    for key, val in json_changes["detailed_changes"]["bioentities"]["by_type"]["cluster"].items():
        text_report += "\n" + key + "\t" + str(val)

    text_report += "\n\nCHANGES IN ANNOTATED BIOENTITIES BY TYPE (ALL)"
    for key, val in json_changes["detailed_changes"]["bioentities"]["by_type"]["all"].items():
        text_report += "\n" + key + "\t" + str(val)

    text_report += "\n\nCHANGES IN ANNOTATED BIOENTITIES BY FILTERED TAXON AND BY TYPE (CLUSTER)"
    text_report += "\ntaxon"
    for type, nb in json_changes["detailed_changes"]["bioentities"]["by_type"]["cluster"].items():
        text_report += "\t" + type
    for key, val in json_changes["detailed_changes"]["bioentities"]["by_taxon"]["cluster"].items():
        text_report += "\n" + key
        for type, nb in json_changes["detailed_changes"]["bioentities"]["by_type"]["cluster"].items():
            text_report += "\t" + str(val[type]["A"]) if type in val else "\t0"

    text_report += "\n\nCHANGES IN ANNOTATED BIOENTITIES BY FILTERED TAXON AND BY TYPE (ALL)"
    text_report += "\ntaxon"
    for type, nb in json_changes["detailed_changes"]["bioentities"]["by_type"]["all"].items():
        text_report += "\t" + type
    for key, val in json_changes["detailed_changes"]["bioentities"]["by_taxon"]["all"].items():
        text_report += "\n" + key
        for type, nb in json_changes["detailed_changes"]["bioentities"]["by_type"]["all"].items():
            text_report += "\t" + str(val[type]["A"]) if type in val else "\t0"


    text_report += "\n\nCHANGES IN ANNOTATED TAXA\n"
    text_report += "total\t" + str(json_changes["detailed_changes"]["taxa"]["total"])
    text_report += "\nfiltered\t" + str(json_changes["detailed_changes"]["taxa"]["filtered"]) + " (with more than 1000 annotations)"


    text_report += "\n\nCHANGES IN ANNOTATIONS\n"
    text_report += "total\t" + str(json_changes["detailed_changes"]["total"])
    for key, val in json_changes["detailed_changes"]["by_aspect"].items():
        text_report += "\n" + key + "\t" + str(val)
        
    text_report += "\n\nCHANGES IN ANNOTATIONS BY BIOENTITY TYPE (CLUSTER)"
    for key, val in json_changes["detailed_changes"]["by_bioentity_type"]["cluster"].items():
        text_report += "\n" + key + "\t" + str(val)
    
    text_report += "\n\nCHANGES IN ANNOTATIONS BY BIOENTITY TYPE (ALL)"
    for key, val in json_changes["detailed_changes"]["by_bioentity_type"]["all"].items():
        text_report += "\n" + key + "\t" + str(val)
    
    text_report += "\n\nCHANGES IN ANNOTATIONS BY EVIDENCE (CLUSTER)"
    for key, val in json_changes["detailed_changes"]["by_evidence"]["cluster"].items():
        text_report += "\n" + key + "\t" + str(val)
    
    text_report += "\n\nCHANGES IN ANNOTATIONS BY EVIDENCE (ALL)"
    for key, val in json_changes["detailed_changes"]["by_evidence"]["all"].items():
        text_report += "\n" + key + "\t" + str(val)
    
    text_report += "\n\nCHANGES IN ANNOTATIONS BY GROUP"
    for key, val in json_changes["detailed_changes"]["by_group"].items():
        text_report += "\n" + key + "\t" + str(val)
    
    text_report += "\n\nCHANGES IN ANNOTATIONS BY TAXON"
    for key, val in json_changes["detailed_changes"]["by_taxon"].items():
        text_report += "\n" + key + "\t" + str(val)


    text_report += "\n\nCHANGES IN REFERENCES AND PMIDS\n"
    text_report += "total\t" + str(json_changes["detailed_changes"]["references"]["all"]["total"])
    text_report += "\t" + str(json_changes["detailed_changes"]["references"]["pmids"]["total"])
 
    text_report += "\n\nCHANGES IN REFERENCES AND PMIDS BY GROUP"
    text_report += "\ngroup\treferences\tpmids"
    for key, val in json_changes["detailed_changes"]["references"]["all"]["by_group"].items():
        text_report += "\n" + key + "\t" + str(val) + "\t" + str(json_changes["detailed_changes"]["references"]["pmids"]["by_group"][key])

    text_report += "\n\nCHANGES IN REFERENCES AND PMIDS BY TAXON"
    text_report += "\ntaxon\treferences\tpmids"
    for key, val in json_changes["detailed_changes"]["references"]["all"]["by_taxon"].items():
        pmid_val = json_changes["detailed_changes"]["references"]["pmids"]["by_taxon"][key] if key in json_changes["detailed_changes"]["references"]["pmids"]["by_taxon"] else 0
        text_report += "\n" + key + "\t" + str(val) + "\t" + str(pmid_val)

    return text_report


def write_json(key, content):
    with open(key, 'w') as outfile:
        json.dump(content, outfile, indent=2)
 
def write_text(key, content):
    with open(key, 'w') as outfile:
        outfile.write(content)


def print_help():
    print('\nUsage: python go_annotation_changes.py -c <current_stats_url> -p <previous_stats_url> -o <output_rep>\n')


def main(argv):
    current_stats_url = ''
    previous_stats_url = ''
    output_rep = ''

    if len(argv) < 3:
        print_help()
        sys.exit(2)

    try:
        opts, argv = getopt.getopt(argv,"c:p:o:",["current=","previous=","orep="])
    except getopt.GetoptError:
        print_help()
        sys.exit(2)

    for opt, arg in opts:
        if opt == '-h':
            print_help()
            sys.exit()
        elif opt in ("-c", "--current"):
            current_stats_url = arg
        elif opt in ("-p", "--previous"):
            previous_stats_url = arg
        elif opt in ("-o", "-orep"):
            output_rep = arg
        
    if not output_rep.endswith("/"):
        output_rep += "/"

    if not os.path.exists(output_rep):
        os.mkdir(output_rep)

    output_json =  output_rep + "go-stats-changes.json"
    output_tsv =  output_rep + "go-stats-changes.tsv"

    print("Will write stats changes to " + output_json + " and " + output_tsv)


    current_stats = requests.get(current_stats_url).json()
    previous_stats = requests.get(previous_stats_url).json()
    json_changes = compute_changes(current_stats, previous_stats)

    print("Saving Stats to <" + output_json + "> ...")    
    write_json(output_json, json_changes)
    print("Done.")

    print("Saving Stats to <" + output_tsv + "> ...")    
    tsv_changes = create_text_report(json_changes)
    write_text(output_tsv, tsv_changes)
    print("Done.")
    


if __name__ == "__main__":
   main(sys.argv[1:])


from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='go-stats',
    version='1.1.3',
    description='Python Library to generate statistics on a Gene Ontology (GO) release',
    py_modules=["go_stats"],
    packages=[''],
    url="https://github.com/geneontology/go-stats",
    author="Laurent-Philippe Albou",
    author_email="laurent.albou@lbl.gov",
    keywords=["GO", "Gene Ontology", "annotation", "ontology", "stats", "changes", "GOLR", "statistics"],
    install_requires=[
        'requests',
        'networkx'
    ],

    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: BSD License"
    ],

    long_description=long_description,
    long_description_content_type="text/markdown"
)


